import unittest

import numpy as np

from pycram.local_transformer import LocalTransformer
from pycram.datastructures.pose import PoseStamped, TransformStamped
from pycram.testing import BulletWorldTestCase
from pycram.ros import Time


class TestLocalTransformer(BulletWorldTestCase):

    def test_singelton(self):
        l1 = LocalTransformer()
        l1.update_transforms([TransformStamped()])
        l1_transforms = l1.get_all_frames()
        l2 = LocalTransformer()
        l2_transforms = l2.get_all_frames()
        self.assertTrue(l1 is l2)
        self.assertTrue(l1_transforms == l2_transforms)

    def test_transform_pose(self):
        l = LocalTransformer()
        l.update_transforms([TransformStamped.from_list([1, 1, 1], [0, 0, 0, 1], "map", "test_frame")])

        p = PoseStamped()
        transformed_pose = l.transform_pose(p, "test_frame")

        self.assertEqual([-1, -1, -1], transformed_pose.position.to_list())
        self.assertEqual([0, 0, 0, 1], transformed_pose.orientation.to_list())
        self.assertEqual("test_frame", transformed_pose.header.frame_id)

    def test_transform_pose_position(self):
        l = LocalTransformer()
        l.update_transforms([TransformStamped.from_list([1, 1, 1], [0, 0, 0, 1], "map", "test_frame")])

        p = PoseStamped()
        transformed_pose = l.transform_pose(p, "test_frame")

        self.assertTrue(transformed_pose.position == transformed_pose.pose.position)

    def test_transform_pose_rotation(self):
        l = LocalTransformer()
        l.update_transforms([TransformStamped.from_list([1, 0, 0], [0, 0, 1, 1], "map", "test_frame")])
        test_pose = PoseStamped.from_list([1, 0, 0], [0, 0, 0, 1], frame="test_frame")
        transformed_pose = l.transform_pose(test_pose, "map")
        np.testing.assert_almost_equal([1, 1, 0], transformed_pose.position.to_list(), decimal=3)


    @unittest.skip
    def test_update_for_object(self):
        l = LocalTransformer()
        self.milk.set_pose(PoseStamped.from_list([1, 2, 1]))
        self.milk.update_link_transforms()
        self.assertTrue(l.can_transform("map", self.milk.tf_frame, Time(0)))

    def test_transform_object_frame(self):
        p = PoseStamped()
        self.milk.set_pose(PoseStamped.from_list([1, 2, 1]))
        lt = LocalTransformer()
        new_pose = lt.transform_to_object_frame(p, self.milk)

        self.assertEqual([-1, -2, -1], new_pose.position.to_list())

    def test_link_wrt(self):
        relative = self.robot.links["torso_lift_link"].get_pose_wrt_link(self.robot.links["torso_lift_link"])
        self.assertEqual([0, 0, 0], relative.position.to_list())
        self.assertEqual([0, 0, 0, 1], relative.orientation.to_list())

        relative = self.robot.links["torso_lift_link"].get_pose_wrt_link(self.robot.links["base_link"])
        for excepted, actual in zip([-0.05, 0, 0.73967], relative.position.to_list()):
            self.assertAlmostEqual(excepted, actual, 4)
        self.assertEqual([0, 0, 0, 1], relative.orientation.to_list())

    def test_transform_to_base(self):
        pose = PoseStamped.from_list([1, 2, 1])
        lt = LocalTransformer()
        base_pose = lt.transform_pose(pose, self.robot.tf_frame)

        self.assertEqual([1,2,1], base_pose.position.to_list())
        self.assertEqual([0, 0, 0, 1], base_pose.orientation.to_list())
        self.assertEqual(self.robot.tf_frame, base_pose.header.frame_id)

    def test_transform_to_base_movement(self):
        pose = PoseStamped.from_list([1, 2, 1])
        lt = LocalTransformer()
        base_pose = lt.transform_pose(pose, self.robot.tf_frame)

        self.assertEqual([1, 2, 1], base_pose.position.to_list())
        self.assertEqual([0, 0, 0, 1], base_pose.orientation.to_list())
        self.assertEqual(self.robot.tf_frame, base_pose.header.frame_id)

        self.robot.pose = PoseStamped.from_list([1, 2, 1])
        base_pose = lt.transform_pose(pose, self.robot.tf_frame)

        self.assertEqual([0, 0, 0], base_pose.position.to_list())
        self.assertEqual([0, 0, 0, 1], base_pose.pose.orientation.to_list())

    def test_transform_base_movement(self):
        pose = PoseStamped.from_list([0, 0, 0])
        lt = LocalTransformer()
        gripper_pose = lt.transform_pose(pose, self.robot.links["l_gripper_tool_frame"].tf_frame)
        np.testing.assert_almost_equal([-0.951, -0.188, -0.7906], gripper_pose.position.to_list(), decimal=3)
        self.assertEqual([0, 0, 0, 1], gripper_pose.orientation.to_list())
        self.assertEqual(self.robot.links["l_gripper_tool_frame"].tf_frame, gripper_pose.header.frame_id)

        self.robot.pose = PoseStamped.from_list([1, 2, 1])
        gripper_pose = lt.transform_pose(pose, self.robot.links["l_gripper_tool_frame"].tf_frame)
        np.testing.assert_almost_equal([-1.951, -2.188, -1.7906], gripper_pose.position.to_list(), decimal=4)
        self.assertEqual([0, 0, 0, 1], gripper_pose.orientation.to_list())
        self.assertEqual(self.robot.links["l_gripper_tool_frame"].tf_frame, gripper_pose.header.frame_id)

    def test_tcp_transform(self):
        tcp_pose = PoseStamped.from_list(frame=self.robot.links["l_gripper_tool_frame"].tf_frame)
        lt = LocalTransformer()
        base_pose = lt.transform_pose(tcp_pose, self.robot.tf_frame)

        np.testing.assert_almost_equal([0.951, 0.188, 0.79067], base_pose.position.to_list(), decimal=5)
        self.assertEqual([0, 0, 0, 1], base_pose.orientation.to_list())
        self.assertEqual(self.robot.tf_frame, base_pose.header.frame_id)

    def test_tcp_transform_movement(self):
        tcp_pose = PoseStamped.from_list([0, 0, 0], frame=self.robot.links["l_gripper_tool_frame"].tf_frame)
        lt = LocalTransformer()
        base_pose = lt.transform_pose(tcp_pose,  self.robot.tf_frame)

        np.testing.assert_almost_equal([0.951, 0.188, 0.79067], base_pose.position.to_list(), decimal=5)
        self.assertEqual([0, 0, 0, 1], base_pose.orientation.to_list())
        self.assertEqual(self.robot.tf_frame, base_pose.header.frame_id)

        self.robot.pose = PoseStamped.from_list([1, 2, 1])
        base_pose = lt.transform_pose(tcp_pose, self.robot.tf_frame)
        np.testing.assert_almost_equal([0.951, 0.188, 0.79067], base_pose.position.to_list(), decimal=5)
        self.assertEqual([0, 0, 0, 1], base_pose.orientation.to_list())
        self.assertEqual(self.robot.tf_frame, base_pose.header.frame_id)

    @unittest.skip("Skip until further investigation is done.")
    def test_tcp_transform_rotation_movement(self):
        tcp_pose = PoseStamped.from_list([0, 0, 0], frame=self.robot.links["l_gripper_tool_frame"].tf_frame)
        lt = LocalTransformer()
        base_pose = lt.transform_pose(tcp_pose, self.robot.tf_frame)

        np.testing.assert_almost_equal([0.951, 0.188, 0.79067], base_pose.position.to_list(), decimal=5)
        self.assertEqual([0, 0, 0, 1], base_pose.orientation.to_list())
        self.assertEqual(self.robot.tf_frame, base_pose.header.frame_id)

        self.robot.pose = PoseStamped.from_list([1, 2, 1], [0, 0, 1, 1])

        robot_base_pose = lt.transform_pose(self.robot.pose, self.robot.tf_frame)
        np.testing.assert_almost_equal(robot_base_pose.position.to_list(), [0,0, 0])
        np.testing.assert_almost_equal(robot_base_pose.orientation.to_list(), [0,0, 0, 1])

        base_pose = lt.transform_pose(tcp_pose, self.robot.tf_frame)
        np.testing.assert_almost_equal([0.951, 0.188, 0.79067], base_pose.position.to_list(), decimal=5)
        np.testing.assert_almost_equal([0, 0, 0, 1], base_pose.orientation.to_list(), decimal=5)
        self.assertEqual(self.robot.tf_frame, base_pose.header.frame_id)

    def test_tcp_transform_rotation_movement_minimal(self):
        base_pose = TransformStamped.from_list([0, 0, 0], [0, 0, 0, 1], "map", "base")
        tcp_pose = TransformStamped.from_list([0.5, 0.5, 1], [0, 0, 0, 1], "map", "tcp")

        lt = LocalTransformer()
        lt.update_transforms([base_pose, tcp_pose])

        test_pose = PoseStamped.from_list([0.1, 0, 0], [0, 0, 0, 1], frame="tcp")
        transformed_pose = lt.transform_pose(test_pose, "base")
        np.testing.assert_almost_equal([0.6, 0.5, 1], transformed_pose.position.to_list(), decimal=3)
        np.testing.assert_almost_equal([0, 0, 0, 1], transformed_pose.orientation.to_list(), decimal=3)

        base_pose = TransformStamped.from_list([0, 0, 0], [0, 0, 1, 1], "map", "base")
        tcp_pose = TransformStamped.from_list([0.5, 0.5, 1], [0, 0, 1, 1], "map", "tcp")
        lt.update_transforms([base_pose, tcp_pose])

        test_pose = PoseStamped.from_list([0.1, 0, 0], [0, 0, 0, 1], frame="tcp")
        transformed_pose = lt.transform_pose(test_pose, "base")
        np.testing.assert_almost_equal([0.6, 0.5, 1], transformed_pose.position.to_list(), decimal=3)
        np.testing.assert_almost_equal([0, 0, 0, 1], transformed_pose.orientation.to_list(), decimal=3)


from __future__ import annotations

import dataclasses

import owlready2
from owlready2.triplelite import _SearchList
from typing_extensions import TYPE_CHECKING, Iterable, Iterator, Union

from ..datastructures.enums import ObjectType
from ..datastructures.partial_designator import PartialDesignator
from ..datastructures.world import World
from ..external_interfaces.robokudo import *
from ..plan import Plan
from ..utils import is_iterable
from ..world_concepts.world_object import Object as WorldObject, Object
from ..description import ObjectDescription

if TYPE_CHECKING:
    pass


class OntologyObjectDesignatorDescription:
    """
    Description for Objects that can be found using ontological reasoning
    """

    search_result: List
    """
    The result from the search in the ontology.
    """

    def __init__(self, search_result: _SearchList):
        self.search_result = list(search_result)

    def __iter__(self) -> Iterable[Object]:
        """
        :return: The objects in the current world which match the search result in the 'is_a' relation.
        """
        for obj in World.current_world.objects:

            # expand is_a of the object individual
            is_a = obj.ontology_individual.is_a + [obj.ontology_individual]

            # get the matching concepts
            intersection = ([x for x in is_a if x in self.search_result])

            # if it matches
            if len(intersection) > 0:
                yield obj


class BelieveObject(ObjectDesignatorDescription):
    """
    Description for Objects that are only believed in.
    """

class ResolutionStrategyObject(ObjectDesignatorDescription):

    def __init__(self, strategy: Union[Callable, Iterable]):
        """
        Description for Objects that are only believed in.

        :param strategy: The strategy to use for the resolution
        """
        super().__init__()
        self.strategy = strategy

    def create_iterator(self, resolution_strategy: Union[Callable, Iterable]):
        """
        Creates an iterator for the given method. If the method is iterable it will be used as is, otherwise it will
        be called as a function.

        :param resolution_strategy: The method to create an iterator for.
        :return: An iterator for the given method.
        """

        class IterClass:
            def __init__(self, method: Union[Callable, Iterable]):
                self.method = method

            def __iter__(self):
                if callable(self.method):
                    yield self.method()
                elif is_iterable(self.method()):
                    for i in self.method():
                        yield i

        if isinstance(resolution_strategy, Plan):
            resolution_strategy = resolution_strategy.perform
        resolution_strategy = IterClass(resolution_strategy)
        return resolution_strategy

    def __iter__(self) -> Iterable[Object]:
        """
        Iterates through every possible solution for the given solution strategy.

        :return: A resolved object designator
        """
        for obj in self.create_iterator(self.strategy):
            yield obj


class ObjectPart(ObjectDesignatorDescription, Iterable[ObjectDescription.Link]):
    """
    Object Designator Descriptions for Objects that are part of some other object.
    """

    def __init__(self, names: List[str],
                 part_of: Union[WorldObject, Iterable[WorldObject]],
                 type: Optional[ObjectType] = None):
        """
        Describing the relationship between an object and a specific part of it.

        :param names: Possible names for the part
        :param part_of: Parent object of which the part should be described
        :param type: Type of the part
        """
        super().__init__(names, type)
        PartialDesignator.__init__(self, ObjectPart, names=names, part_of=part_of, type=type)

        if not part_of:
            raise AttributeError("part_of cannot be None.")

        self.type: Optional[ObjectType] = type
        self.names: Optional[List[str]] = names
        self.part_of = part_of

    def ground(self) -> ObjectDescription.Link:
        """
        Default specialized_designators, returns the first result of the iterator of this instance.

        :return: A resolved object designator
        """
        return next(iter(self))

    def __iter__(self) -> Iterator[ObjectDescription.Link]:
        """
        Iterates through every possible solution for the given input parameter.

        :yield: A resolved Object designator
        """
        for params in self.generate_permutations():
            if params["names"] in params["part_of"].links.keys():
                yield params["part_of"].links[params["names"]]



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Optional, Union, List

from .enums import Grasp, AxisIdentifier
from ..has_parameters import HasParameters, has_parameters


@has_parameters
@dataclass
class GraspDescription(HasParameters):
    """
    Represents a grasp description with a side grasp, top face, and orientation alignment.
    """

    approach_direction: Grasp
    """
    The primary approach direction. Must be one of {Grasp.FRONT, Grasp.BACK, Grasp.LEFT, Grasp.RIGHT}.
    """

    vertical_alignment: Optional[Grasp] = None
    """
    The vertical alignment when grasping the pose, or None if not applicable. Must be one of {Grasp.TOP, Grasp.BOTTOM, None}.
    """

    rotate_gripper: bool = False
    """
    Indicates if the gripper should be rotated by 90°. Must be a boolean.
    """

    def __post_init__(self):
        allowed_approach_direction = {Grasp.FRONT, Grasp.BACK, Grasp.LEFT, Grasp.RIGHT}
        if self.approach_direction not in allowed_approach_direction:
            raise ValueError(
                f"Invalid value for side_face: {self.approach_direction}. Allowed values are {allowed_approach_direction}")
        allowed_vertical_alignment = {Grasp.TOP, Grasp.BOTTOM, None}
        if self.vertical_alignment not in allowed_vertical_alignment:
            raise ValueError(
                f"Invalid value for top_face: {self.vertical_alignment}. Allowed values are {allowed_vertical_alignment}")

    def __hash__(self):
        return hash((self.approach_direction, self.vertical_alignment, self.rotate_gripper))

    def as_list(self) -> List[Union[Grasp, Optional[Grasp], bool]]:
        """
        :return: A list representation of the grasp description.
        """
        return [self.approach_direction, self.vertical_alignment, self.rotate_gripper]


@dataclass(frozen=True)
class PreferredGraspAlignment:
    """
    Description of the preferred grasp alignment for an object.
    """
    preferred_axis: Optional[AxisIdentifier]
    """
    The preferred axis, X, Y, or Z, for grasping the object, or None if not specified.
    """

    with_vertical_alignment: bool
    """
    Indicates if the object should be grasped with a vertical alignment.
    """

    with_rotated_gripper: bool
    """
    Indicates if the gripper should be rotated by 90° around X.
    """



import os

from setuptools import setup, find_packages

current_directory = os.path.dirname(os.path.abspath(__file__))


# Function to read the version from random_events.__version__
def get_version():
    version_file = os.path.join(current_directory, 'src', 'pycram', '__init__.py')
    with open(version_file, 'r') as f:
        for line in f:
            if line.startswith('__version__'):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


# Optional project description in README.md:
try:
    with open(os.path.join(current_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
except Exception:
    long_description = ''


# Iterate through all the files and subdirectories
# to build the data files array
def generate_data_files(share_path, dir):
    data_files = []
    for path, _, files in os.walk(dir):
        list_entry = (share_path + path, [os.path.join(path, f) for f in files if not f.startswith('.')])
        data_files.append(list_entry)
    return data_files

if os.environ.get('ROS_VERSION') == "1":
    ## ! DO NOT MANUALLY INVOKE THIS setup.py, USE CATKIN INSTEAD

    from distutils.core import setup
    from catkin_pkg.python_setup import generate_distutils_setup

    # fetch values from package.xml
    setup_args = generate_distutils_setup(
        packages=['pycram'],
        package_dir={'': 'src'}
    )

    setup(**setup_args)
elif os.environ.get('ROS_VERSION') == "2":

    package_name = 'pycram'

    setup(
        name=package_name,
        version=get_version(),
        # packages=find_packages(exclude=['test'], include=['pycram', 'pycrap']),
        package_dir={"": "src"},  # Optional
        packages=["pycram", "pycrap"],
        data_files=[
                       ('share/ament_index/resource_index/packages',
                        ['resource/' + package_name]),
                       ('share/' + package_name, ['package.xml']),
                   ] + generate_data_files('share/' + package_name + '/', 'resources'),
        install_requires=['setuptools', 'pycram_bullet', 'numpy', 'pytransform3d', 'typing_extensions>=4.10.0',
                          'trimesh==4.6.0', 'random_events>=4.1.0', 'pin==2.7.0', 'transforms3d', 'python-box',
                          'urdf_parser_py', 'networkx', 'pint', 'owlready2>=0.45', 'psutil', 'deprecated', "plotly", "sortedcontainers", "probabilistic_model", "manifold3d"],
        zip_safe=True,
        maintainer='Jonas Dech',
        maintainer_email='jdech@uni-bremen.de',
        description='A Python library for cognitive robot control',
        long_description=long_description,
        long_description_content_type='text/markdown',
        license='Gplv3',
        tests_require=['pytest'],
        entry_points={
            'console_scripts': [
            ],
        },
    )

else:
    setup(
        name="pycram-robotics",
        version=get_version(),
        packages=find_packages(where='src'),
        package_dir={"": "src"},  # Optional
        data_files=generate_data_files("lib/python3.10/site-packages/pycram/", 'resources'),
        # packages=["pycram", "pycrap"],
        install_requires=['setuptools', 'pycram_bullet', 'numpy', 'pytransform3d', 'typing_extensions>=4.10.0',
                          'trimesh==4.6.0', 'random_events>=4.1.0', 'pin==2.7.0', 'transforms3d', 'python-box',
                          'urdf_parser_py', 'networkx', 'pint', 'owlready2>=0.45', 'psutil', 'deprecated', "plotly", "sortedcontainers", "probabilistic_model", "manifold3d"],
        zip_safe=True,
        maintainer='Jonas Dech',
        maintainer_email='jdech@uni-bremen.de',
        description='A Python library for cognitive robot control',
        long_description=long_description,
        long_description_content_type='text/markdown',
        license='Gplv3',
        tests_require=['pytest'],
        entry_points={
            'console_scripts': [
            ],
        },
    )



import unittest
import os

from pycram.external_interfaces.procthor import ProcThorInterface
from pycram.ros import ros_tools
import socket
import shutil


def is_server_reachable(host, port):
    """Check if a server is reachable by attempting to open a socket connection."""
    try:
        with socket.create_connection((host, port), timeout=5):
            return True
    except (socket.timeout, OSError):
        return False


class TestProcThorInterface(unittest.TestCase):
    @unittest.skipIf(not(is_server_reachable("procthor.informatik.uni-bremen.de", 5000)),"Host unreachable")
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.test_folder_path = os.path.join(ros_tools.get_ros_package_path('pycram') + "/tmp")
        cls.procThorInterface = ProcThorInterface(base_url="http://procthor.informatik.uni-bremen.de:5000/",
                                                  source_folder=cls.test_folder_path)

    def setUp(self) -> None:
        super().setUp()
        if not os.path.exists(self.test_folder_path):
            os.mkdir(self.test_folder_path)

    def tearDown(self) -> None:
        if os.path.exists(self.test_folder_path):
            shutil.rmtree(self.test_folder_path)

    @classmethod
    def TearDownClass(cls):
        super().TearDownClass()


    def test_download_one(self):
        self.procThorInterface.download_one_random_environment()
        found_dir=[name for name in os.listdir(self.test_folder_path) if os.path.isdir(os.path.join(self.test_folder_path,name))]
        self.assertTrue(1==len(found_dir))

    def test_download_multiple(self):
        self.procThorInterface.download_num_random_environment(3)
        found_dir = [name for name in os.listdir(self.test_folder_path) if
                     os.path.isdir(os.path.join(self.test_folder_path, name))]
        self.assertTrue(3 == len(found_dir))

    def test_get_all_environments_stored_below_dictionary(self):
        before=self.procThorInterface.get_all_environments_stored_below_directory(self.test_folder_path)
        self.procThorInterface.download_one_random_environment()
        after=self.procThorInterface.get_all_environments_stored_below_directory(self.test_folder_path)
        self.assertTrue(1==len(after)-len(before))


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import os
from pathlib import Path

import numpy as np
from deprecated import deprecated
from trimesh.parent import Geometry3D
from typing_extensions import Type, Optional, Dict, Tuple, List, Union, Any

from ..datastructures.dataclasses import (Color, ObjectState, LinkState, JointState,
                                          AxisAlignedBoundingBox, VisualShape, ClosestPointsList,
                                          ContactPointsList, RotatedBoundingBox, VirtualJoint, FrozenObject, FrozenLink,
                                          FrozenJoint)
from ..datastructures.enums import ObjectType, JointType
from ..datastructures.pose import PoseStamped, TransformStamped, Point, Quaternion, Vector3
from ..datastructures.world import World
from ..datastructures.world_entity import PhysicalBody
from ..description import ObjectDescription, LinkDescription, Joint
from ..failures import ObjectAlreadyExists, WorldMismatchErrorBetweenAttachedObjects, UnsupportedFileExtension, \
    ObjectDescriptionUndefined
from ..has_parameters import HasParameters, leaf_types
from ..local_transformer import LocalTransformer
from ..object_descriptors.generic import ObjectDescription as GenericObjectDescription
from ..object_descriptors.urdf import ObjectDescription as URDF
from ..ros import logwarn, logerr, Time

try:
    from ..object_descriptors.mjcf import ObjectDescription as MJCF
except ImportError:
    MJCF = None
from ..robot_description import RobotDescriptionManager, RobotDescription
from ..world_concepts.constraints import Attachment
from pycrap.ontologies import PhysicalObject, Joint, \
    Robot, Floor, Location, Bowl, Spoon, Cereal, Environment

Link = ObjectDescription.Link


class Object(PhysicalBody, HasParameters):
    """
    Represents a spawned Object in the World.
    """

    tf_prospection_world_prefix: str = "prospection/"
    """
    The prefix for the tf frame of objects in the prospection world.
    """

    extension_to_description_type: Dict[str, Type[ObjectDescription]] = {URDF.get_file_extension(): URDF}
    """
    A dictionary that maps the file extension to the corresponding ObjectDescription type.
    """

    ontology_concept: Type[PhysicalObject] = PhysicalObject

    def __init__(self, name: str, concept: Type[PhysicalObject], path: Optional[str] = None,
                 description: Optional[ObjectDescription] = None,
                 pose: Optional[PoseStamped] = None,
                 world: Optional[World] = None,
                 color: Optional[Color] = None,
                 size: Optional[List[float]] = None,
                 ignore_cached_files: bool = False,
                 scale_mesh: Optional[float] = None,
                 mesh_transform: Optional[TransformStamped] = None):
        """
        The constructor loads the description file into the given World, if no World is specified the
        :py:attr:`~World.current_world` will be used. It is also possible to load .obj and .stl file into the World.
        The rgba_color parameter is only used when loading .stl or .obj files,
        for URDFs :func:`~Object.set_color` can be used.

        :param name: The name of the object
        :param concept: The type of the object as ontological concept from PyCRAP
        :param path: The path to the source file, if only a filename is provided then the resources directories will be
         searched, it could be None in some cases when for example it is a generic object.
        :param description: The ObjectDescription of the object, this contains the joints and links of the object.
        :param pose: The pose at which the Object should be spawned
        :param world: The World in which the object should be spawned, if no world is specified the
         :py:attr:`~World.current_world` will be used.
        :param color: The rgba_color with which the object should be spawned.
        :param size: The dimensions of the object in float as a list.
        :param ignore_cached_files: If true the file will be spawned while ignoring cached files.
        :param scale_mesh: The scale of the mesh.
        """

        self.world = world if world is not None else World.current_world
        self.name: str = name
        super().__init__(-1, self.world, concept=concept)

        pose = PoseStamped() if pose is None else pose

        self.path: Optional[str] = path

        self._resolve_description(path, description)
        self.cache_manager = self.world.cache_manager

        self.local_transformer = LocalTransformer()
        self.original_pose = self.local_transformer.transform_pose(pose, "map")
        self._current_pose = self.original_pose
        self.scale_mesh = scale_mesh if scale_mesh is not None else 1.0
        color = Color() if color is None else color

        if path is not None:
            self.path = self.world.preprocess_object_file_and_get_its_cache_path(path, ignore_cached_files,
                                                                                 self.description, self.name,
                                                                                 scale_mesh=self.scale_mesh,
                                                                                 mesh_transform=mesh_transform,
                                                                                 color=color)

            self.description.update_description_from_file(self.path)
        # if the object is an agent in the belief state
        if self.is_a_robot and not self.world.is_prospection_world:
            self._update_world_robot_and_description()

        # if the object is an environment in the belief state
        if self.is_an_environment and not self.world.is_prospection_world:
            self._update_world_environment_object()

        self.id = self._spawn_object_and_get_id()

        self._init_joint_name_and_id_map()
        self._init_link_name_and_id_map()

        self._init_links_and_update_transforms()

        if color is not None:
            self.color: Color = color

        self.size_ = size

        self._init_joints()

        self.attachments: Dict[Object, Attachment] = {}

        self.world.add_object(self)

    @property
    def parts(self) -> Dict[str, PhysicalBody]:
        return self.links

    @property
    def tf_frame(self) -> str:
        """
        The tf frame of the object.
        """
        return (self.tf_prospection_world_prefix if self.world.is_prospection_world else "") + self.name

    @property
    def color(self) -> Union[Color, Dict[str, Color]]:
        """
        Return the rgba_color of this object. The return is either:

            1. A Color object with RGBA values, this is the case if the object only has one link (this
                happens for example if the object is spawned from a .obj or .stl file)
            2. A dict with the link name as key and the rgba_color as value. The rgba_color is given as a Color Object.
                Please keep in mind that not every link may have a rgba_color. This is dependent on the URDF from which
                 the object is spawned.

        :return: The rgba_color as Color object with RGBA values between 0 and 1 or a dict with the link name as key and
         the rgba_color as value.
        """
        link_to_color_dict = self.links_colors

        if len(link_to_color_dict) == 1:
            return list(link_to_color_dict.values())[0]
        else:
            return link_to_color_dict

    @color.setter
    def color(self, rgba_color: Union[Color, Dict[str, Color]]) -> None:
        """
        Change the color of this object.

        :param rgba_color: The color as Color object with RGBA values between 0 and 1
        """
        # Check if there is only one link, this is the case for primitive
        # forms or if loaded from an .stl or .obj file
        if self.has_one_link:
            self.root_link.color = rgba_color
        else:
            if isinstance(rgba_color, Color):
                for link in self.links.values():
                    link.color = rgba_color
            else:
                for link_name, color in rgba_color.items():
                    self.links[link_name].color = color

    @property
    def size(self) -> List[float]:
        return self.size_

    def get_mesh_path(self) -> List[str]:
        """
        Get the path to the mesh file of the object.

        :return: The path(s) to the mesh file(s).
        """
        if self.has_one_link:
            return self.root_link.get_mesh_path()
        else:
            raise ValueError("The object has more than one link, therefore the mesh path cannot be determined.")

    def _resolve_description(self, path: Optional[str] = None, description: Optional[ObjectDescription] = None) -> None:
        """
        Find the correct description type of the object and initialize it and set the description of this object to it.

        :param path: The path to the source file.
        :param description: The ObjectDescription of the object.
        """
        if description is not None:
            self.description = description
            return
        if path is None:
            raise ObjectDescriptionUndefined(self.name)
        extension = Path(path).suffix
        if extension in self.extension_to_description_type:
            self.description = self.extension_to_description_type[extension]()
        elif extension in ObjectDescription.mesh_extensions:
            self.description = self.world.conf.default_description_type()
        else:
            raise UnsupportedFileExtension(self.name, path)

    def set_mobile_robot_pose(self, pose: PoseStamped) -> None:
        """
        Set the goal for the mobile base joints of a mobile robot to reach a target pose. This is used for example when
        the simulator does not support setting the pose of the robot directly (e.g. MuJoCo).

        :param pose: The target pose.
        """
        goal = self.get_mobile_base_joint_goal(pose)
        goal = {vj.name: pos for vj, pos in goal.items()}
        self.set_multiple_joint_positions(goal)

    def get_mobile_base_joint_goal(self, pose: PoseStamped) -> Dict[VirtualJoint, float]:
        """
        Get the goal for the mobile base joints of a mobile robot to reach a target pose.

        :param pose: The target pose.
        :return: The goal for the mobile base joints.
        """
        # target_translation, target_angle = self.get_mobile_base_pose_difference(pose)
        # Get the joints of the base link
        mobile_base_joints = self.world.get_robot_mobile_base_joints()
        return {mobile_base_joints.translation_x: pose.position.x,
                mobile_base_joints.translation_y: pose.position.y,
                mobile_base_joints.angular_z: pose.z_angle}

    def get_mobile_base_pose_difference(self, pose: PoseStamped) -> Tuple[Point, float]:
        """
        Get the difference between the current and the target pose of the mobile base.

        :param pose: The target pose.
        :return: The difference between the current and the target pose of the mobile base.
        """
        return self.original_pose.get_position_diff(pose), self.original_pose.get_z_angle_difference(pose)

    @property
    def joint_actuators(self) -> Optional[Dict[str, str]]:
        """
        The joint actuators of the robot.
        """
        if self.obj_type == ObjectType.ROBOT:
            return self.robot_description.joint_actuators
        return None

    @property
    def has_actuators(self) -> bool:
        """
        True if the object has actuators, otherwise False.
        """
        return self.robot_description.has_actuators

    @property
    def robot_description(self) -> RobotDescription:
        """
        The current robot description.
        """
        return self.world.robot_description

    def get_actuator_for_joint(self, joint: Joint) -> Optional[str]:
        """
        Get the actuator name for a joint.

        :param joint: The joint object for which to get the actuator.
        :return: The name of the actuator.
        """
        return self.robot_description.get_actuator_for_joint(joint.name)

    def get_multiple_link_positions(self, links: List[Link]) -> Dict[str, List[float]]:
        """
        Get the positions of multiple links of the object.

        :param links: The link objects of which to get the positions.
        :return: The positions of the links.
        """
        return self.world.get_multiple_link_positions(links)

    def get_multiple_link_orientations(self, links: List[Link]) -> Dict[str, List[float]]:
        """
        Get the orientations of multiple links of the object.

        :param links: The link objects of which to get the orientations.
        :return: The orientations of the links.
        """
        return self.world.get_multiple_link_orientations(links)

    def get_multiple_link_poses(self, links: List[Link]) -> Dict[str, PoseStamped]:
        """
        Get the poses of multiple links of the object.

        :param links: The link objects of which to get the poses.
        :return: The poses of the links.
        """
        return self.world.get_multiple_link_poses(links)

    def get_poses_of_attached_objects(self) -> Dict[Object, PoseStamped]:
        """
        Get the poses of the attached objects.

        :return: The poses of the attached objects
        """
        return {child_object: attachment.get_child_object_pose()
                for child_object, attachment in self.attachments.items() if not attachment.loose}

    def get_target_poses_of_attached_objects_given_parent(self, pose: PoseStamped) -> Dict[Object, PoseStamped]:
        """
        Get the target poses of the attached objects of an object. Given the pose of the parent object. (i.e. the poses
         to which the attached objects will move when the parent object is at the given pose)

        :param pose: The pose of the parent object.
        :return: The target poses of the attached objects
        """
        return {child_object: attachment.get_child_object_pose_given_parent(pose) for child_object, attachment
                in self.attachments.items() if not attachment.loose}

    @property
    def name(self):
        """
        The name of the object.
        """
        return self._name

    @name.setter
    def name(self, name: str):
        """
        Set the name of the object.
        """
        self._name = name
        if name in [obj.name for obj in self.world.objects]:
            raise ObjectAlreadyExists(self)

    @property
    def pose(self):
        """
        The current pose of the object.
        """
        return self.world.get_object_pose(self)

    @pose.setter
    def pose(self, pose: PoseStamped):
        """
        Set the pose of the object.
        """
        self.set_pose(pose)

    @property
    def transform(self):
        """
        The current transform of the object.
        """
        return self.get_pose().to_transform_stamped(self.tf_frame)

    @property
    def obj_type(self) -> Type[PhysicalObject]:
        return self.ontology_concept

    def _spawn_object_and_get_id(self) -> int:
        """
        Loads an object to the given World with the given position and orientation. The rgba_color will only be
        used when an .obj or .stl file is given.
        If a .obj or .stl file is given, before spawning, an urdf file with the .obj or .stl as mesh will be created
        and this URDf file will be loaded instead.
        When spawning a URDf file a new file will be created in the cache directory, if there exists none.
        This new file will have resolved mesh file paths, meaning there will be no references
        to ROS packages instead there will be absolute file paths.

        :return: The unique id of the object and the path of the file that was loaded.
        """
        if isinstance(self.description, GenericObjectDescription):
            return self.world.load_generic_object_and_get_id(self.description, pose=self._current_pose)

        path = self.path if self.world.conf.let_pycram_handle_spawning else self.name

        try:
            obj_id = self.world.load_object_and_get_id(path, self._current_pose, self.obj_type)
            return obj_id

        except Exception as e:
            logerr(
                f"The caught error: {e}, The File could not be loaded. Please note that the path has to be either"
                f" a URDF, stl or obj file or the name of an URDF string on the parameter server.")
            os.remove(path)
            raise e

    def _update_world_robot_and_description(self):
        """
        Initialize the robot description of the object, load the description from the RobotDescriptionManager and set
        the robot as the current robot in the World. Also add the virtual mobile base joints to the robot.
        """
        rdm = RobotDescriptionManager()
        rdm.load_description(self.description.name)
        World.robot = self
        self._add_virtual_move_base_joints()

    def _update_world_environment_object(self):
        """
        Initialize the environment as the current environment in the World.
        """
        World.environment = self

    def _add_virtual_move_base_joints(self):
        """
        Add the virtual mobile base joints to the robot description.
        """
        virtual_joints = self.robot_description.virtual_mobile_base_joints
        if virtual_joints is None:
            return
        child_link = self.description.get_root()
        axes = virtual_joints.get_axes()
        for joint_name, joint_type in virtual_joints.get_types().items():
            self.description.add_joint(joint_name, child_link, joint_type, axes[joint_name], is_virtual=True)

    def _init_joint_name_and_id_map(self) -> None:
        """
        Create a dictionary which maps the joint names to their unique ids and vice versa.
        """
        n_joints = len(self.joint_names)
        self.joint_name_to_id = dict(zip(self.joint_names, range(n_joints)))
        self.joint_id_to_name = dict(zip(self.joint_name_to_id.values(), self.joint_name_to_id.keys()))

    def _init_link_name_and_id_map(self) -> None:
        """
        Create a dictionary which maps the link names to their unique ids and vice versa.
        """
        n_links = len(self.link_names)
        self.link_name_to_id: Dict[str, int] = dict(zip(self.link_names, range(n_links)))
        self.link_name_to_id[self.description.get_root()] = -1
        self.link_id_to_name: Dict[int, str] = dict(zip(self.link_name_to_id.values(), self.link_name_to_id.keys()))

    def _init_links_and_update_transforms(self) -> None:
        """
        Initialize the link objects from the URDF file and creates a dictionary which maps the link names to the
        corresponding link objects.
        """
        self.links = {}
        for link_name, link_id in self.link_name_to_id.items():
            link_description = self.description.get_link_by_name(link_name)
            if link_name == self.description.get_root():
                self.links[link_name] = self.description.RootLink(self)
            else:
                self.links[link_name] = self.description.Link(link_id, link_description, self)

        self.update_link_transforms()

    def _init_joints(self):
        """
        Initialize the joint objects from the URDF file and creates a dictionary which mas the joint names to the
        corresponding joint objects
        """
        self.joints = {}
        for joint_name, joint_id in self.joint_name_to_id.items():
            parsed_joint_description = self.description.get_joint_by_name(joint_name)
            is_virtual = self.is_joint_virtual(joint_name)
            self.joints[joint_name] = self.description.Joint(joint_id, parsed_joint_description, self,
                                                             is_virtual=is_virtual)

    def is_joint_virtual(self, name: str):
        """
        Check if a joint is virtual.
        """
        return self.description.is_joint_virtual(name)

    @property
    def virtual_joint_names(self):
        """
        The names of the virtual joints.
        """
        return self.description.virtual_joint_names

    @property
    def virtual_joints(self):
        """
        The virtual joints as a list.
        """
        return [joint for joint in self.joints.values() if joint.is_virtual]

    @property
    def has_one_link(self) -> bool:
        """
        True if the object has only one link, otherwise False.
        """
        return len(self.links) == 1

    @property
    def link_names(self) -> List[str]:
        """
        The names of the links as a list.
        """
        return self.world.get_object_link_names(self)

    @property
    def joint_names(self) -> List[str]:
        """
        The names of the joints as a list.
        """
        joint_names = self.world.get_object_joint_names(self)
        return joint_names

    def get_link(self, link_name: str) -> ObjectDescription.Link:
        """
        Return the link object with the given name.

        :param link_name: The name of the link.
        :return: The link object.
        """
        return self.links[link_name]

    def get_link_pose(self, link_name: str) -> PoseStamped:
        """
        Return the pose of the link with the given name.

        :param link_name: The name of the link.
        :return: The pose of the link.
        """
        return self.links[link_name].pose

    def get_link_position(self, link_name: str) -> Point:
        """
        Return the position of the link with the given name.

        :param link_name: The name of the link.
        :return: The position of the link.
        """
        return self.links[link_name].position

    def get_link_position_as_list(self, link_name: str) -> List[float]:
        """
        Return the position of the link with the given name.

        :param link_name: The name of the link.
        :return: The position of the link.
        """
        return self.links[link_name].position.to_list()

    def get_link_orientation(self, link_name: str) -> Quaternion:
        """
        Return the orientation of the link with the given name.

        :param link_name: The name of the link.
        :return: The orientation of the link.
        """
        return self.links[link_name].orientation

    def get_link_orientation_as_list(self, link_name: str) -> List[float]:
        """
        Return the orientation of the link with the given name.

        :param link_name: The name of the link.
        :return: The orientation of the link.
        """
        return self.links[link_name].orientation.to_list()

    def get_link_tf_frame(self, link_name: str) -> str:
        """
        Return the tf frame of the link with the given name.

        :param link_name: The name of the link.
        :return: The tf frame of the link.
        """
        return self.links[link_name].tf_frame

    def get_link_axis_aligned_bounding_box(self, link_name: str,
                                           transform_to_link_pose: bool = True) -> AxisAlignedBoundingBox:
        """
        Return the axis aligned bounding box of the link with the given name.

        :param link_name: The name of the link.
        :param transform_to_link_pose: If True, the bounding box will be transformed to fit link pose.
        :return: The axis aligned bounding box of the link.
        """
        return self.links[link_name].get_axis_aligned_bounding_box(transform_to_link_pose)

    def get_transform_between_links(self, from_link: str, to_link: str) -> TransformStamped:
        """
        Return the transform between two links.

        :param from_link: The name of the link from which the transform should be calculated.
        :param to_link: The name of the link to which the transform should be calculated.
        """
        return self.links[from_link].get_transform_to_link(self.links[to_link])

    def get_link_color(self, link_name: str) -> Color:
        """
        Return the color of the link with the given name.

        :param link_name: The name of the link.
        :return: The color of the link.
        """
        return self.links[link_name].color

    def set_link_color(self, link_name: str, color: List[float]) -> None:
        """
        Set the color of the link with the given name.

        :param link_name: The name of the link.
        :param color: The new color of the link.
        """
        self.links[link_name].color = Color.from_list(color)

    def get_link_geometry(self, link_name: str) -> List[VisualShape]:
        """
        Return the collision geometry of the link with the given name.

        :param link_name: The name of the link.
        :return: List of the collision geometry of the link.
        """
        return self.links[link_name].geometry

    def get_link_visual_geometry(self, link_name: str) -> List[VisualShape]:
        """
        Return the visual geometry of the link with the given name.

        :param link_name: The name of the link.
        :return: The visual geometry of the link.
        """
        return self.links[link_name].visual_geometry

    def get_link_transform(self, link_name: str) -> TransformStamped:
        """
        Return the transform of the link with the given name.

        :param link_name: The name of the link.
        :return: The transform of the link.
        """
        return self.links[link_name].transform

    def get_link_origin(self, link_name: str) -> PoseStamped:
        """
        Return the origin of the link with the given name.

        :param link_name: The name of the link.
        :return: The origin of the link as a 'Pose'.
        """
        return self.links[link_name].origin

    def get_link_origin_transform(self, link_name: str) -> TransformStamped:
        """
        Return the origin transform of the link with the given name.

        :param link_name: The name of the link.
        :return: The origin transform of the link.
        """
        return self.links[link_name].origin_transform

    @property
    def base_origin_shift(self) -> np.ndarray:
        """
        The shift between the base of the object and the origin of the object.

        :return: A numpy array with the shift between the base of the object and the origin of the object.
        """
        return np.array(self.get_position_as_list()) - np.array(self.get_base_position_as_list())

    def __repr__(self):
        return self.__class__.__qualname__ + (f"(name={self.name}, object_type={self.obj_type.name},"
                                              f" file_path={self.path}, pose={self.pose}, world={self.world})")

    def remove(self) -> None:
        """
        Remove this object from the World it currently resides in.
        For the object to be removed it has to be detached from all objects it
        is currently attached to. Then remove this Object from the simulation/world.
        """
        self.world.remove_object(self)

    def reset(self, remove_saved_states=False) -> None:
        """
        Reset the Object to the state it was first spawned in.
        All attached objects will be detached, all joints will be set to the
        default position of 0 and the object will be set to the position and
        orientation in which it was spawned.

        :param remove_saved_states: If True the saved states will be removed.
        """
        self.reset_concepts()
        self.detach_all()
        self.reset_all_links()
        self.reset_all_joints()
        self.set_pose(self.original_pose)
        if remove_saved_states:
            self.remove_saved_states()

    def reset_all_joints(self) -> None:
        """
        Reset all joints of the object.
        """
        for joint in self.joints.values():
            joint.reset_concepts()
        self.reset_all_joints_positions()

    def reset_all_links(self) -> None:
        """
        Reset all links of the object.
        """
        for link in self.links.values():
            link.reset()

    @property
    def is_an_environment(self) -> bool:
        """
        Check if the object is of type environment.

        :return: True if the object is of type environment, False otherwise.
        """
        return (issubclass(self.obj_type, Location) or issubclass(self.obj_type, Floor)
                or issubclass(self.obj_type, Environment))

    @property
    def is_an_object(self) -> bool:
        """
        Check if the object is of type Physical Object or Generic Object.

        :return: True if the object is of type PhysicalObject , False otherwise.
        """
        return issubclass(self.obj_type, PhysicalObject)

    @property
    def is_a_robot(self) -> bool:
        """
        Check if the object is a robot.
        TODO: Check if this is a the correct filter
        :return: True if the object is a robot, False otherwise.
        """
        return issubclass(self.obj_type, Robot)

    def merge(self, other: Object, name: Optional[str] = None, pose: Optional[PoseStamped] = None,
              new_description_file: Optional[str] = None) -> Object:
        """
        Merge the object with another object. This is done by merging the descriptions of the objects,
        removing the original objects creating a new merged object.

        :param other: The object to merge with.
        :param name: The name of the merged object.
        :param pose: The pose of the merged object.
        :param new_description_file: The new description file of the merged object.
        :return: The merged object.
        """
        pose = self.pose if pose is None else pose
        child_pose = self.local_transformer.transform_pose(other.pose, self.tf_frame)
        description = self.description.merge_description(other.description, child_pose_wrt_parent=child_pose,
                                                         new_description_file=new_description_file)
        name = self.name if name is None else name
        color = self.color if isinstance(self.color, Color) else self.color[self.root_link.name]
        other.remove()
        self.remove()
        return Object(name, self.obj_type, description.xml_path, description=description, pose=pose, world=self.world,
                      color=color)

    def attach(self,
               child_object: Object,
               parent_link: Optional[str] = None,
               child_link: Optional[str] = None,
               bidirectional: bool = True,
               coincide_the_objects: bool = False,
               parent_to_child_transform: Optional[TransformStamped] = None) -> None:
        """
        Attach another object to this object. This is done by
        saving the transformation between the given link, if there is one, and
        the base pose of the other object. Additionally, the name of the link, to
        which the object is attached, will be saved.
        Furthermore, a simulator constraint will be created so the attachment
        also works while simulation.
        Loose attachments means that the attachment will only be one-directional. For example, if this object moves the
        other, attached, object will also move but not the other way around.

        :param child_object: The other object that should be attached.
        :param parent_link: The link name of this object.
        :param child_link: The link name of the other object.
        :param bidirectional: If the attachment should be a loose attachment.
        :param coincide_the_objects: If True the object frames will be coincided.
        :param parent_to_child_transform: The transform from the parent to the child object.
        """
        parent_link = self.links[parent_link] if parent_link else self.root_link
        child_link = child_object.links[child_link] if child_link else child_object.root_link

        if coincide_the_objects and parent_to_child_transform is None:
            parent_to_child_transform = TransformStamped.from_list()
        attachment = Attachment(parent_link, child_link, bidirectional, parent_to_child_transform)

        self.attachments[child_object] = attachment
        child_object.attachments[self] = attachment.get_inverse()

        self.world.attachment_event(self, [self, child_object])

    def detach(self, child_object: Object) -> None:
        """
        Detache another object from this object. This is done by
        deleting the attachment from the attachments dictionary of both objects
        and deleting the constraint of the simulator.
        Afterward the detachment event of the corresponding World will be fired.

        :param child_object: The object which should be detached
        """
        del self.attachments[child_object]
        del child_object.attachments[self]

        self.world.detachment_event(self, [self, child_object])

    def detach_all(self) -> None:
        """
        Detach all objects attached to this object.
        """
        attachments = self.attachments.copy()
        for att in attachments.keys():
            self.detach(att)

    def update_attachment_with_object(self, child_object: Object):
        self.attachments[child_object].update_transform_and_constraint()

    def get_position(self) -> Point:
        """
        Return the position of this Object as a list of xyz.

        :return: The current position of this object
        """
        return self.get_pose().position

    def get_orientation(self) -> PoseStamped.orientation:
        """
        Return the orientation of this object as a list of xyzw, representing a quaternion.

        :return: A list of xyzw
        """
        return self.get_pose().orientation

    def get_position_as_list(self) -> List[float]:
        """
        Return the position of this Object as a list of xyz.

        :return: The current position of this object
        """
        return self.get_pose().position.to_list()

    def get_base_position_as_list(self) -> List[float]:
        """
        Return the position of this Object as a list of xyz.

        :return: The current position of this object
        """
        return self.get_base_origin().position.to_list()

    def get_orientation_as_list(self) -> List[float]:
        """
        Return the orientation of this object as a list of xyzw, representing a quaternion.

        :return: A list of xyzw
        """
        return self.get_pose().orientation.to_list()

    def get_pose(self) -> PoseStamped:
        """
        Return the position of this object as a list of xyz. Alias for :func:`~Object.get_position`.

        :return: The current pose of this object
        """
        return self.pose

    def set_pose(self, pose: PoseStamped, base: bool = False, set_attachments: bool = True) -> None:
        """
        Set the Pose of the object.

        :param pose: New Pose for the object
        :param base: If True places the object base instead of origin at the specified position and orientation
        :param set_attachments: Whether to set the poses of the attached objects to this object or not.
        """
        pose_in_map = self.local_transformer.transform_pose(pose, "map")
        if base:
            pose_in_map.position = Vector3.from_list(np.array(pose_in_map.position.to_list()) + self.base_origin_shift)

        self.reset_base_pose(pose_in_map)

        if set_attachments:
            self._set_attached_objects_poses()

    def reset_base_pose(self, pose: PoseStamped) -> bool:
        return self.world.reset_object_base_pose(self, pose)

    def move_base_to_origin_pose(self) -> None:
        """
        Move the object such that its base will be at the current origin position.
        This is useful when placing objects on surfaces where you want the object base in contact with the surface.
        """
        self.set_pose(self.get_pose(), base=True)

    def save_state(self, state_id: int, save_dir: Optional[str] = None) -> None:
        """
        Save the state of this object by saving the state of all links and attachments.

        :param state_id: The unique id of the state.
        :param save_dir: The directory in which to save the state.
        """
        self.save_links_states(state_id)
        self.save_joints_states(state_id)
        super().save_state(state_id, save_dir)

    def save_links_states(self, state_id: int) -> None:
        """
        Save the state of all links of this object.

        :param state_id: The unique id of the state.
        """
        for link in self.links.values():
            link.save_state(state_id)

    def save_joints_states(self, state_id: int) -> None:
        """
        Save the state of all joints of this object.

        :param state_id: The unique id of the state.
        """
        for joint in self.joints.values():
            joint.save_state(state_id)

    @property
    def current_state(self) -> ObjectState:
        """
        The current state of this object as an ObjectState.
        """
        return ObjectState(self.body_state, self.attachments.copy(), self.link_states.copy(),
                           self.joint_states.copy())

    @current_state.setter
    def current_state(self, state: ObjectState) -> None:
        """
        Set the current state of this object to the given state.
        """
        if self.current_state != state:
            self.body_state = state.body_state
            self.set_attachments(state.attachments)
            self.link_states = state.link_states
            self.joint_states = state.joint_states

    def set_attachments(self, attachments: Dict[Object, Attachment]) -> None:
        """
        Set the attachments of this object to the given attachments.

        :param attachments: A dictionary with the object as key and the attachment as value.
        """
        self.detach_objects_not_in_attachments(attachments)
        self.attach_objects_in_attachments(attachments)

    def detach_objects_not_in_attachments(self, attachments: Dict[Object, Attachment]) -> None:
        """
        Detach objects that are not in the attachments list and are in the current attachments list.

        :param attachments: A dictionary with the object as key and the attachment as value.
        """
        copy_of_attachments = self.attachments.copy()
        for obj, attachment in copy_of_attachments.items():
            original_obj = obj
            if self.world.is_prospection_world and len(attachments) > 0 \
                    and not list(attachments.keys())[0].world.is_prospection_world:
                obj = self.world.get_object_for_prospection_object(obj)
            if obj not in attachments:
                if attachment.is_inverse:
                    original_obj.detach(self)
                else:
                    self.detach(original_obj)

    def attach_objects_in_attachments(self, attachments: Dict[Object, Attachment]) -> None:
        """
        Attach objects that are in the given attachments list but not in the current attachments list.

        :param attachments: A dictionary with the object as key and the attachment as value.
        """
        for obj, attachment in attachments.items():
            is_prospection = self.world.is_prospection_world and not obj.world.is_prospection_world
            if is_prospection:
                obj = self.world.get_prospection_object_for_object(obj)
            if obj in self.attachments:
                if self.attachments[obj] != attachment:
                    if attachment.is_inverse:
                        obj.detach(self)
                    else:
                        self.detach(obj)
                else:
                    continue
            self.mimic_attachment_with_object(attachment, obj)

    def mimic_attachment_with_object(self, attachment: Attachment, child_object: Object) -> None:
        """
        Mimic the given attachment for this and the given child objects.

        :param attachment: The attachment to mimic.
        :param child_object: The child object.
        """
        att_transform = self.get_attachment_transform_with_object(attachment, child_object)
        if attachment.is_inverse:
            child_object.attach(self, attachment.child_link.name, attachment.parent_link.name,
                                attachment.bidirectional,
                                parent_to_child_transform=~att_transform)
        else:
            self.attach(child_object, attachment.parent_link.name, attachment.child_link.name,
                        attachment.bidirectional, parent_to_child_transform=att_transform)

    def get_attachment_transform_with_object(self, attachment: Attachment, child_object: Object) -> TransformStamped:
        """
        Return the attachment transform for the given parent and child objects, taking into account the prospection
        world.

        :param attachment: The attachment.
        :param child_object: The child object.
        :return: The attachment transform.
        """
        if self.world != child_object.world:
            raise WorldMismatchErrorBetweenAttachedObjects(self, child_object)
        att_transform = attachment.parent_to_child_transform.copy()
        if self.world.is_prospection_world and not attachment.parent_object.world.is_prospection_world:
            att_transform.frame_id = self.tf_prospection_world_prefix + att_transform.frame_id
            att_transform.child_frame_id = self.tf_prospection_world_prefix + att_transform.child_frame_id
        return att_transform

    @property
    def link_states(self) -> Dict[int, LinkState]:
        """
        The current state of all links of this object.

        :return: A dictionary with the link id as key and the current state of the link as value.
        """
        return {link.id: link.current_state for link in self.links.values()}

    @link_states.setter
    def link_states(self, link_states: Dict[int, LinkState]) -> None:
        """
        Set the current state of all links of this object.

        :param link_states: A dictionary with the link id as key and the current state of the link as value.
        """
        for link in self.links.values():
            link.current_state = link_states[link.id]

    @property
    def joint_states(self) -> Dict[int, JointState]:
        """
        The current state of all joints of this object.

        :return: A dictionary with the joint id as key and the current state of the joint as value.
        """
        return {joint.id: joint.current_state for joint in self.joints.values()}

    @joint_states.setter
    def joint_states(self, joint_states: Dict[int, JointState]) -> None:
        """
        Set the current state of all joints of this object.

        :param joint_states: A dictionary with the joint id as key and the current state of the joint as value.
        """
        for joint in self.joints.values():
            if joint.name not in self.robot_virtual_move_base_joints_names():
                joint.current_state = joint_states[joint.id]

    def robot_virtual_move_base_joints_names(self):
        if self.robot_description.virtual_mobile_base_joints is None:
            return []
        return self.robot_description.virtual_mobile_base_joints.names

    def remove_saved_states(self) -> None:
        """
        Remove all saved states of this object.
        """
        super().remove_saved_states()
        self.remove_links_saved_states()
        self.remove_joints_saved_states()

    def remove_links_saved_states(self) -> None:
        """
        Remove all saved states of the links of this object.
        """
        for link in self.links.values():
            link.remove_saved_states()

    def remove_joints_saved_states(self) -> None:
        """
        Remove all saved states of the joints of this object.
        """
        for joint in self.joints.values():
            joint.remove_saved_states()

    def _set_attached_objects_poses(self, already_moved_objects: Optional[List[Object]] = None) -> None:
        """
        Update the positions of all attached objects. This is done
        by calculating the new pose in world coordinate frame and setting the
        base pose of the attached objects to this new pose.
        After this call _set_attached_objects method for all attached objects.

        :param already_moved_objects: A list of Objects that were already moved, these will be excluded to prevent loops
         in the update.
        """
        if not self.world.conf.let_pycram_move_attached_objects:
            return

        if already_moved_objects is None:
            already_moved_objects = []

        for child in self.attachments:

            if child in already_moved_objects:
                continue

            attachment = self.attachments[child]
            if attachment.loose:
                self.update_attachment_with_object(child)
                child.update_attachment_with_object(self)

            else:
                child.set_pose(attachment.get_child_link_target_pose(), set_attachments=False)
                child._set_attached_objects_poses(already_moved_objects + [self])

    def set_position(self, position: Union[PoseStamped, Point, List], base=False) -> None:
        """
        Set this Object to the given position, if base is true, place the bottom of the Object at the position
        instead of the origin in the center of the Object. The given position can either be a Pose,
        in this case only the position is used or a geometry_msgs.msg/Point which is the position part of a Pose.

        :param position: Target position as xyz.
        :param base: If the bottom of the Object should be placed or the origin in the center.
        """
        pose = PoseStamped()
        if isinstance(position, PoseStamped):
            target_position = position.position
            pose.frame_id = position.frame_id
        elif isinstance(position, Point):
            target_position = position
        elif isinstance(position, (List, np.ndarray, tuple)):
            if len(position) == 3:
                target_position = Point(**dict(zip(["x", "y", "z"], position)))
            else:
                raise ValueError("The given position has to be a sequence of 3 values.")
        else:
            raise TypeError("The given position has to be a Pose, Point or a sequence of xyz values.")

        pose.position = target_position
        pose.orientation = self.get_orientation()
        self.set_pose(pose, base=base)

    def set_orientation(self, orientation: Union[PoseStamped, Quaternion, List, Tuple, np.ndarray]) -> None:
        """
        Set the orientation of the Object to the given orientation. Orientation can either be a Pose, in this case only
        the orientation of this pose is used or a geometry_msgs.msg/Quaternion which is the orientation of a Pose.

        :param orientation: Target orientation given as a list of xyzw.
        """
        pose = PoseStamped()
        if isinstance(orientation, PoseStamped):
            target_orientation = orientation.orientation
            pose.frame_id = orientation.frame_id
        elif isinstance(orientation, Quaternion):
            target_orientation = orientation
        elif (isinstance(orientation, list) or isinstance(orientation, np.ndarray) or isinstance(orientation, tuple)) \
                and len(orientation) == 4:
            target_orientation = Quaternion(**dict(zip(["x", "y", "z", "w"], orientation)))
        else:
            raise TypeError("The given orientation has to be a Pose, Quaternion or one of list/tuple/ndarray of xyzw.")

        pose.pose.position = self.get_position()
        pose.pose.orientation = target_orientation
        self.set_pose(pose)

    def get_joint_id(self, name: str) -> int:
        """
        Return the unique id for a joint name. As used by the world/simulator.

        :param name: The joint name
        :return: The unique id
        """
        return self.joint_name_to_id[name]

    def get_root_link_description(self) -> LinkDescription:
        """
        Return the root link of the URDF of this object.

        :return: The root link as defined in the URDF of this object.
        """
        for link_description in self.description.links:
            if link_description.name == self.description.get_root():
                return link_description

    @property
    def root_link(self) -> ObjectDescription.Link:
        """
        The root link of this object.

        :return: The root link of this object.
        """
        return self.links[self.description.get_root()]

    @property
    def tip_link(self) -> ObjectDescription.Link:
        """
        The tip link of this object.

        :return: The tip link of this object.
        """
        return self.links[self.description.get_tip()]

    def get_root_link_id(self) -> int:
        """
        Return the unique id of the root link of this object.

        :return: The unique id of the root link of this object.
        """
        return self.get_link_id(self.description.get_root())

    def get_link_id(self, link_name: str) -> int:
        """
        Return a unique id for a link name.

        :param link_name: The name of the link.
        :return: The unique id of the link.
        """
        return self.link_name_to_id[link_name]

    def get_link_by_id(self, link_id: int) -> ObjectDescription.Link:
        """
        Return the link for a given unique link id

        :param link_id: The unique id of the link.
        :return: The link object.
        """
        return self.links[self.link_id_to_name[link_id]]

    def reset_all_joints_positions(self) -> None:
        """
        Set the current position of all joints to 0. This is useful if the joints should be reset to their default
        """
        joint_names = [joint.name for joint in self.joints.values()]
        if len(joint_names) == 0:
            return
        joint_positions = [0] * len(joint_names)
        self.set_multiple_joint_positions(dict(zip(joint_names, joint_positions)))

    def set_joint_position(self, joint_name: str, joint_position: float) -> None:
        """
        Set the position of the given joint to the given joint pose and updates the poses of all attached objects.

        :param joint_name: The name of the joint
        :param joint_position: The target pose for this joint
        """
        self.clip_joint_positions_to_limits({joint_name: joint_position})
        if self.world.reset_joint_position(self.joints[joint_name], joint_position):
            self._set_attached_objects_poses()

    @deprecated("Use set_multiple_joint_positions instead")
    def set_joint_positions(self, joint_positions: Dict[str, float]) -> None:
        self.set_multiple_joint_positions(joint_positions)

    def set_multiple_joint_positions(self, joint_positions: Dict[str, float]) -> None:
        """
        Set the current position of multiple joints at once, this method should be preferred when setting
        multiple joints at once instead of running :func:`~Object.set_joint_position` in a loop.

        :param joint_positions: A dictionary with the joint names as keys and the target positions as values.
        """
        joint_positions = {self.joints[joint_name]: joint_position
                           for joint_name, joint_position in joint_positions.items()}
        if self.world.set_multiple_joint_positions(joint_positions):
            self._set_attached_objects_poses()

    def clip_joint_positions_to_limits(self, joint_positions: Dict[str, float]) -> Dict[str, float]:
        """
        Clip the given joint positions to the joint limits.

        :param joint_positions: A dictionary with the joint names as keys and the target positions as values.
        :return: A dictionary with the joint names as keys and the clipped positions as values.
        """
        return {joint_name: np.clip(joint_position, self.joints[joint_name].lower_limit,
                                    self.joints[joint_name].upper_limit)
        if self.joints[joint_name].has_limits else joint_position
                for joint_name, joint_position in joint_positions.items()}

    def get_joint_position(self, joint_name: str) -> float:
        """
        Return the current position of the given joint.

        :param joint_name: The name of the joint
        :return: The current position of the given joint
        """
        return self.joints[joint_name].position

    def get_joint_damping(self, joint_name: str) -> float:
        """
        Return the damping of the given joint (friction).

        :param joint_name: The name of the joint
        :return: The damping of the given joint
        """
        return self.joints[joint_name].damping

    def get_joint_upper_limit(self, joint_name: str) -> float:
        """
        Return the upper limit of the given joint.

        :param joint_name: The name of the joint
        :return: The upper limit of the given joint
        """
        return self.joints[joint_name].upper_limit

    def get_joint_lower_limit(self, joint_name: str) -> float:
        """
        Return the lower limit of the given joint.

        :param joint_name: The name of the joint
        :return: The lower limit of the given joint
        """
        return self.joints[joint_name].lower_limit

    def get_joint_axis(self, joint_name: str) -> Point:
        """
        Return the axis of the given joint.

        :param joint_name: The name of the joint
        :return: The axis of the given joint
        """
        return self.joints[joint_name].axis

    def get_joint_type(self, joint_name: str) -> JointType:
        """
        Return the type of the given joint.

        :param joint_name: The name of the joint
        :return: The type of the given joint
        """
        return self.joints[joint_name].type

    def get_joint_limits(self, joint_name: str) -> Tuple[float, float]:
        """
        Return the lower and upper limits of the given joint.

        :param joint_name: The name of the joint
        :return: The lower and upper limits of the given joint
        """
        return self.joints[joint_name].limits

    def get_joint_child_link(self, joint_name: str) -> ObjectDescription.Link:
        """
        Return the child link of the given joint.

        :param joint_name: The name of the joint
        :return: The child link of the given joint
        """
        return self.joints[joint_name].child_link

    def get_joint_parent_link(self, joint_name: str) -> ObjectDescription.Link:
        """
        Return the parent link of the given joint.

        :param joint_name: The name of the joint
        :return: The parent link of the given joint
        """
        return self.joints[joint_name].parent_link

    def find_joint_above_link(self, link_name: str, joint_type: Optional[JointType] = None) -> Optional[str]:
        """
        Traverses the chain from 'link' to the URDF origin and returns the first joint that is of type 'joint_type'.
        If no joint type is given, the first joint that is not FIXED is returned.

        :param link_name: AbstractLink name above which the joint should be found
        :param joint_type: Joint type that should be searched for
        :return: Name of the first joint which has the given type
        """
        chain = self.description.get_chain(self.description.get_root(), link_name)
        reversed_chain = reversed(chain)
        for element in reversed_chain:
            if element not in self.joint_name_to_id:
                continue

            element_joint_type = self.get_joint_type(element)
            if joint_type is not None and element_joint_type == joint_type:
                return element

            if joint_type is None and element_joint_type != JointType.FIXED:
                return element

        logwarn(f"No joint of type {joint_type} found above link {link_name}")
        return None

    def get_multiple_joint_positions(self, joint_names: List[str]) -> Dict[str, float]:
        """
        Return the positions of multiple joints at once.

        :param joint_names: A list of joint names.
        :return: A dictionary with the joint names as keys and the joint positions as values.
        """
        return self.world.get_multiple_joint_positions([self.joints[joint_name] for joint_name in joint_names])

    def get_positions_of_controllable_joints(self) -> Dict[str, float]:
        """
        Return a list of all controllable joints of this object.

        :return: A list of all controllable joints.
        """
        return {j.name: j.position for j in self.joints.values()
                if j.type != JointType.FIXED and not j.is_virtual}

    def get_positions_of_all_joints(self) -> Dict[str, float]:
        """
        Return the positions of all joints of the object as a dictionary of joint names and joint positions.

        :return: A dictionary with all joints positions'.
        """
        return {j.name: j.position for j in self.joints.values()}

    def update_link_transforms(self, transform_time: Optional[Time] = None) -> None:
        """
        Update the transforms of all links of this object using time 'transform_time' or the current ros time.

        :param transform_time: The time to use for the transform update.
        """
        for link in self.links.values():
            link.update_transform(transform_time)

    @property
    def contact_points(self) -> ContactPointsList:
        """
        Return a list of contact points of this Object with other Objects.

        :return: A list of all contact points with other objects
        """
        return self.world.get_object_contact_points(self)

    def contact_points_simulated(self) -> ContactPointsList:
        """
        Return a list of all contact points between this Object and other Objects after stepping the simulation once.

        :return: A list of contact points between this Object and other Objects
        """
        state_id = self.world.save_state()
        self.world.step()
        contact_points = self.contact_points
        self.world.restore_state(state_id)
        return contact_points

    def closest_points(self, max_distance: float) -> ClosestPointsList:
        """
        Return a list of closest points between this Object and other Objects.

        :param max_distance: The maximum distance between the closest points
        :return: A list of closest points between this Object and other Objects
        """
        return self.world.get_body_closest_points(self, max_distance)

    def closest_points_with_obj(self, other_object: Object, max_distance: float) -> ClosestPointsList:
        """
        Return a list of closest points between this Object and another Object.

        :param other_object: The other object
        :param max_distance: The maximum distance between the closest points
        :return: A list of closest points between this Object and the other Object
        """
        return self.world.get_closest_points_between_two_bodies(self, other_object, max_distance)

    def set_color(self, rgba_color: Color) -> None:
        self.color = rgba_color

    def get_color(self) -> Union[Color, Dict[str, Color]]:
        return self.color

    @property
    def links_colors(self) -> Dict[str, Color]:
        """
        The color of each link as a dictionary with link names as keys and RGBA colors as values.
        """
        return self.world.get_colors_of_object_links(self)

    def get_axis_aligned_bounding_box(self, shift_to_object_position: bool = True) -> AxisAlignedBoundingBox:
        """
        Return the axis aligned bounding box of this object.

        :param shift_to_object_position: If True, the bounding box will be shifted to the object position.
        :return: The axis aligned bounding box of this object.
        """
        if self.has_one_link:
            return self.root_link.get_axis_aligned_bounding_box(shift_to_object_position)
        else:
            return self.world.get_object_axis_aligned_bounding_box(self)

    def get_rotated_bounding_box(self) -> RotatedBoundingBox:
        """
        Return the rotated bounding box of this object.

        :return: The rotated bounding box of this object.
        """
        if self.has_one_link:
            return self.root_link.get_rotated_bounding_box()
        else:
            return self.world.get_object_rotated_bounding_box(self)

    def get_convex_hull(self) -> Geometry3D:
        """
        Return the convex hull of this object.

        :return: The convex hull of this object.
        """
        if self.has_one_link:
            return self.root_link.get_convex_hull()
        else:
            return self.world.get_body_convex_hull(self)

    def get_base_origin(self) -> PoseStamped:
        """
        Return the origin of the base/bottom of this object.

        :return: the origin of the base/bottom of this object.
        """
        aabb = self.get_axis_aligned_bounding_box()
        base_width = np.absolute(aabb.min_x - aabb.max_x)
        base_length = np.absolute(aabb.min_y - aabb.max_y)
        return PoseStamped.from_list([aabb.min_x + base_width / 2, aabb.min_y + base_length / 2, aabb.min_z],
                                     self.get_orientation_as_list())

    def get_joint_by_id(self, joint_id: int) -> Joint:
        """
        Return the joint object with the given id.

        :param joint_id: The unique id of the joint.
        :return: The joint object.
        """
        return dict([(joint.id, joint) for joint in self.joints.values()])[joint_id]

    def get_link_for_attached_objects(self) -> Dict[Object, ObjectDescription.Link]:
        """
        Return a dictionary which maps attached object to the link of this object to which the given object is attached.

        :return: The link of this object to which the given object is attached.
        """
        return {obj: attachment.parent_link for obj, attachment in self.attachments.items()}

    def copy_to_prospection(self) -> Object:
        """
        Copy this object to the prospection world.

        :return: The copied object in the prospection world.
        """
        return self.copy_to_world(self.world.prospection_world)

    def copy_to_world(self, world: World) -> Object:
        """
        Copy this object to the given world.

        :param world: The world to which the object should be copied.
        :return: The copied object in the given world.
        """
        obj = Object(self.name, self.obj_type, self.path, self.description, self.original_pose,
                     world, self.color)
        return obj

    def parent_entity(self) -> World:
        """
        :return: The parent of this object which is the world.
        """
        return self.world

    def frozen_copy(self) -> FrozenObject:
        """
        Creates a copied version of this object which contains the information of this object but can not be interacted
        with.

        :return FrozenObject: The copied forzen object.
        """
        frozen_links = {l_name: FrozenLink(l.name, l.pose, l.geometry) for l_name, l in self.links.items()}
        frozen_joints = {j_name: FrozenJoint(j.name, j.type, [j.child], j.parent, j.current_state.position) for
                         j_name, j in self.joints.items()}

        return FrozenObject(self.name, self.obj_type, self.path, self.description, self.pose,
                            frozen_links, frozen_joints)

    @classmethod
    def define_parameters(cls) -> Dict[str, Any]:
        """
        Defines the parameters of Object for the HasParameter flattener. Relevant parameters for objects are only the
        Pose and Type.

        :return: A dictionary with the parameters of the object
        """
        params = {}
        params.update(PoseStamped._parameters)
        params["obj_type"] = PhysicalObject
        return params



from pycrap.ontologies.crax.classes import *


class BowlPreferredGraspAlignment(PreferredGraspAlignment):
    ...

class DefaultPreferredGraspAlignment(PreferredGraspAlignment):
    ...

class SpoonPreferredGraspAlignment(PreferredGraspAlignment):
    ...

class CerealPreferredGraspAlignment(PreferredGraspAlignment):
    ...





from __future__ import annotations

import os
import pickle
from abc import ABC, abstractmethod
from copy import copy
from threading import RLock

from trimesh.parent import Geometry3D
from typing_extensions import TYPE_CHECKING, Dict, Optional, List, deprecated, Union, Type, Tuple

from pycrap.ontologies import PhysicalObject, Room, Location, Robot, Floor
from .dataclasses import State, ContactPointsList, ClosestPointsList, Color, PhysicalBodyState, \
    AxisAlignedBoundingBox, RotatedBoundingBox, RayResult
from .enums import AdjacentBodyMethod, AxisIdentifier, Arms, Grasp
from .mixins import HasConcept
from ..local_transformer import LocalTransformer
from ..ros import Time, logdebug
from .pose import GraspDescription, Vector3, PoseStamped
from .grasp import PreferredGraspAlignment

if TYPE_CHECKING:
    from ..datastructures.world import World
    from ..world_concepts.world_object import Object
    from .pose import PoseStamped, GeoQuaternion as Quaternion, TransformStamped, Point


class StateEntity:
    """
    The StateEntity class is used to store the state of an object or the physics simulator. This is used to save and
    restore the state of the World.
    """

    SAVED_FILE_PREFIX = 'pycram_saved_state_'

    def __init__(self):
        self._saved_states: Dict[int, State] = {}

    @property
    def saved_states(self) -> Dict[int, State]:
        """
        :return: the saved states of this entity.
        """
        return self._saved_states

    def save_state(self, state_id: int, save_dir: Optional[str] = None) -> int:
        """
        Saves the state of this entity with the given state id.

        :param state_id: The unique id of the state.
        :param save_dir: The directory where the file should be saved.
        """
        self._saved_states[state_id] = self.current_state
        if save_dir is not None:
            self.save_state_to_file(self.get_saved_file_path_of_state(save_dir, state_id), state_id)
        return state_id

    def save_state_to_file(self, file_path: str, state_id: Optional[int] = None):
        """
        Saves the state of this entity to a file.

        :param file_path: The path to the file.
        :param state_id: The unique id of the state if saving a specific state not the current state.
        """
        state_to_save = self.current_state if state_id is None else self.saved_states[state_id]
        with open(file_path, 'wb') as file:
            pickle.dump(state_to_save, file)

    def restore_state_from_file(self, save_dir: str, state_id: int) -> None:
        """
        Restores the state of this entity from a file.

        :param save_dir: The directory where the file is saved.
        :param state_id: The unique id of the state.
        """
        with open(self.get_saved_file_path_of_state(save_dir, state_id), 'rb') as file:
            self.current_state = pickle.load(file)

    def get_saved_file_path_of_state(self, save_dir: str, state_id: int) -> str:
        """
        Gets the path of the file that stores the state with the given id.

        :param save_dir: The directory where the file should be saved.
        :param state_id: The unique id of the state.
        :return: The name of the file that stores the state with the given id.
        """
        return os.path.join(save_dir, f'{self.SAVED_FILE_PREFIX}{state_id}.pkl')

    @property
    @abstractmethod
    def current_state(self) -> State:
        """
        :return: The current state of this entity.
        """
        pass

    @current_state.setter
    @abstractmethod
    def current_state(self, state: State) -> None:
        """
        Sets the current state of this entity.

        :param state: The new state of this entity.
        """
        pass

    def restore_state(self, state_id: int, saved_file_dir: Optional[str] = None) -> None:
        """
        Restores the state of this entity from a saved state using the given state id.

        :param state_id: The unique id of the state.
        :param saved_file_dir: The directory where the file is saved.
        """
        if saved_file_dir is not None:
            self.restore_state_from_file(saved_file_dir, state_id)
        else:
            self.current_state = self.saved_states[state_id]

    def remove_saved_states(self) -> None:
        """
        Removes all saved states of this entity.
        """
        self._saved_states = {}


class WorldEntity(StateEntity, HasConcept):
    """
    A class that represents an entity of the world, such as an object or a link.
    """

    def __init__(self, _id: int, world: World, concept: Type[PhysicalObject] = PhysicalObject, parse_name: bool = True):
        StateEntity.__init__(self)
        self.id = _id
        self.world: World = world
        HasConcept.__init__(self, name=self.name, world=self.world, concept=concept, parse_name=parse_name)

    def reset_concepts(self):
        """
        Reset the concepts of this entity.
        """
        if not self.world.is_prospection_world and hasattr(self.ontology_individual, 'reload'):
            self.ontology_individual.reload()

    @property
    @abstractmethod
    def name(self) -> str:
        """
        :return: The name of this body.
        """
        ...

    @property
    @abstractmethod
    def parent_entity(self) -> Optional[WorldEntity]:
        """
        :return: The parent entity of this entity, if it has one.
        """
        pass

    def __eq__(self, other: WorldEntity) -> bool:
        """
        Check if this body is equal to another body.
        """
        if not isinstance(other, WorldEntity):
            return False
        return self.id == other.id and self.name == other.name and self.parent_entity == other.parent_entity

    def __hash__(self) -> int:
        return hash((self.id, self.name, self.parent_entity))


class PhysicalBody(WorldEntity):
    """
    A class that represents a physical body in the world that has some related physical properties.
    """

    def __init__(self, body_id: int, world: World, concept: Type[PhysicalObject] = PhysicalObject,
                 parse_name: bool = True):
        WorldEntity.__init__(self, body_id, world, concept, parse_name)

        self.local_transformer = LocalTransformer()
        self._is_translating: Optional[bool] = None
        self._is_rotating: Optional[bool] = None
        self._velocity: Optional[List[float]] = None
        self.ontology_lock: RLock = RLock()
        self.updated_containment_of_parts: bool = False
        self.latest_known_parts: Dict[str, PhysicalBody] = {}

    def reset_concepts(self):
        super().reset_concepts()
        self.contained_bodies = []
        for part in self.parts.values():
            part.reset_concepts()

    @property
    def is_an_environment(self) -> bool:
        """
        Check if the object is of type environment.

        :return: True if the object is of type environment, False otherwise.
        """
        return ((isinstance(self.parent_entity, PhysicalBody) and self.parent_entity.is_an_environment) or
                (issubclass(self.ontology_concept, Location) or issubclass(self.ontology_concept, Floor)))

    @property
    def is_a_robot(self) -> bool:
        """
        Check if the object is a robot.
        TODO: Check if this is a the correct filter
        :return: True if the object is a robot, False otherwise.
        """
        return (issubclass(self.ontology_concept, Robot) or
                (isinstance(self.parent_entity, PhysicalBody) and self.parent_entity.is_a_robot))

    @property
    @abstractmethod
    def parts(self) -> Dict[str, PhysicalBody]:
        """
        :return: The parts of this body as a dictionary mapping the part name to the part.
        """
        ...

    def update_containment(self, excluded_bodies: Optional[List[PhysicalBody]] = None,
                           candidate_selection_method: AdjacentBodyMethod = AdjacentBodyMethod.ClosestPoints,
                           max_distance: float = 0.5) -> None:
        """
        Update the containment of the object by checking if it is contained in other bodies,
         excluding the given excluded bodies.

        :param excluded_bodies: The bodies that should be excluded from the containment check.
        :param candidate_selection_method: The method to select the candidates for the containment check.
        :param max_distance: The maximum distance from this body to other bodies to consider a body as a candidate.
        """
        excluded_bodies = [] if excluded_bodies is None else excluded_bodies
        excluded_bodies.append(self)
        if candidate_selection_method == AdjacentBodyMethod.ClosestPoints:
            bodies = self.get_adjacent_bodies_using_closest_points(max_distance)
        else:
            bodies = self.get_adjacent_bodies_using_rays(max_distance)
        for body in bodies:
            if body in excluded_bodies:
                continue
            if body.get_axis_aligned_bounding_box().contains_box(self.get_axis_aligned_bounding_box()):
                logdebug(f"{body.name} contains {self.name}")
                body.contained_bodies = [self]

    def get_adjacent_bodies_using_closest_points(self, max_distance: float = 0.5) -> List[PhysicalBody]:
        """
        Get the adjacent bodies of the body using the closest points between the bodies.

        :param max_distance: The maximum distance to consider a body as adjacent.
        :return: The bodies that are adjacent to this body if any.
        """
        closest_points = self.closest_points(max_distance)
        return closest_points.get_all_bodies(excluded=[self])

    def get_adjacent_bodies_using_rays(self, max_distance: float = 0.5) -> List[PhysicalBody]:
        """
        Cast rays in the 6 rays in all directions (+x, -x, +y, -y, +z,  and -z directions) to find the adjacent bodies
         of the body.

        :param max_distance: The max distance that the rays can travel.
        :return: The bodies that are adjacent to this body if any.
        """
        if max_distance <= 0:
            raise ValueError("The distance should be greater than zero.")

        rays_start, rays_end = self.cast_rays_in_all_directions(max_distance)
        rays_results = self.world.ray_test_batch(rays_start, rays_end)

        bodies_ids = set([(rr.obj_id, rr.link_id) for rr in rays_results if rr.intersected])

        bodies = [self.world.get_object_by_id(obj_id).get_link_by_id(link_id) for obj_id, link_id in bodies_ids]

        return bodies

    def cast_rays_in_all_directions(self, max_distance: float = 0.5) -> List[RayResult]:
        """
        Get the rays in all 6 directions (+x, -x, +y, -y, +z,  and -z directions).

        :param max_distance: The maximum distance the rays can travel.
        :return: The rays in all 6 directions.
        """
        bb = self.get_axis_aligned_bounding_box()
        origin = bb.origin
        min_, max_ = bb.get_min_max()

        rays_start, rays_end = [], []
        for i in range(3):
            i_rays_start, i_rays_end = self.get_axis_rays(origin, min_[i], max_[i], i, max_distance)
            rays_start.extend(i_rays_start)
            rays_end.extend(i_rays_end)

        return self.world.ray_test_batch(rays_start, rays_end)

    @staticmethod
    def get_axis_rays(origin: List[float], min_val: float, max_val: float, idx: int, max_distance: float = 0.5)\
            -> Tuple[List[List[float]], List[List[float]]]:
        """
        Get the rays in the given axis.

        :param origin: The origin of the rays.
        :param min_val: The start in the -ve direction of the axis.
        :param max_val: The start in the +ve direction of the axis.
        :param idx: The index of the direction/axis.
        :param max_distance: The maximum distance the rays can travel.
        """
        ray_left_start = [min_val if i == idx else origin[i] for i in range(3)]
        ray_left_end = [min_val - max_distance if i == idx else origin[i] for i in range(3)]
        ray_right_start = [max_val if i == idx else origin[i] for i in range(3)]
        ray_right_end = [max_val + max_distance if i == idx else origin[i] for i in range(3)]
        return [ray_left_start, ray_right_start], [ray_left_end, ray_right_end]

    def contains_body(self, body: PhysicalBody) -> bool:
        """
        Check if this body contains another body.

        :param body: The physical body to check if it is contained by this body.
        :return: True if the body contains the other body, otherwise False.
        """
        with self.ontology_lock:
            contained_bodies = self.contained_bodies
            return body in contained_bodies or (body.parent_entity and body.parent_entity in contained_bodies)

    @property
    def contained_bodies(self) -> List[PhysicalBody]:
        """
        :return: True if the object contains the other object, otherwise False.
        """
        with self.ontology_lock:
            self.world.ontology.reason()
            return [self.world.ontology.python_objects[phys_obj]
                    for phys_obj in self.ontology_individual.contains_object]

    @contained_bodies.setter
    def contained_bodies(self, bodies: List[PhysicalBody]) -> None:
        """
        Set the bodies that are contained by this body.

        :param bodies: The bodies that are contained in this body.
        """
        with self.ontology_lock:
            self.ontology_individual.contains_object = [body.ontology_individual for body in bodies]

    @abstractmethod
    def get_axis_aligned_bounding_box(self) -> AxisAlignedBoundingBox:
        """
        :return: The axis-aligned bounding box of this body.
        """
        ...

    @abstractmethod
    def get_rotated_bounding_box(self) -> RotatedBoundingBox:
        """
        :return: The rotated bounding box of this body.
        """
        ...

    def _plot_convex_hull(self):
        """
        Plot the convex hull of the geometry.
        """
        self.get_convex_hull().show()

    @abstractmethod
    def get_convex_hull(self) -> Geometry3D:
        """
        :return: The convex hull of this body.
        """
        ...

    @property
    def body_state(self) -> PhysicalBodyState:
        return PhysicalBodyState(self.pose.copy(), self.is_translating, self.is_rotating, copy(self.velocity)
                                 , self.world.conf.get_pose_tolerance())

    @body_state.setter
    def body_state(self, state: PhysicalBodyState) -> None:
        if self.body_state != state:
            self.pose = state.pose
            self.is_translating = state.is_translating
            self.is_rotating = state.is_rotating
            self.velocity = state.velocity

    @property
    def velocity(self) -> Optional[List[float]]:
        return self._velocity

    @velocity.setter
    def velocity(self, velocity: List[float]) -> None:
        self._velocity = velocity

    @property
    def is_translating(self) -> Optional[bool]:
        return self._is_translating

    @is_translating.setter
    def is_translating(self, is_translating: bool) -> None:
        self._is_translating = is_translating

    @property
    def is_rotating(self) -> Optional[bool]:
        return self._is_rotating

    @is_rotating.setter
    def is_rotating(self, is_rotating: bool) -> None:
        self._is_rotating = is_rotating

    @property
    def position(self) -> Point:
        """
        :return: A Point object containing the position of the link relative to the world frame.
        """
        return self.pose.position


    @property
    def orientation(self) -> Quaternion:
        """
        :return: A Quaternion object containing the orientation of the link relative to the world frame.
        """
        return self.pose.orientation

    @property
    def pose_as_list(self) -> List[List[float]]:
        """
        :return: A list containing the position and orientation of the link relative to the world frame.
        """
        return self.pose.to_list()

    @property
    def transform(self) -> TransformStamped:
        """
        The transform of this entity.

        :return: The transform of this entity.
        """
        return self.pose.to_transform_stamped(self.tf_frame)

    def update_transform(self, transform_time: Optional[Time] = None) -> None:
        """
        Update the transformation of this link at the given time.

        :param transform_time: The time at which the transformation should be updated.
        """
        self.local_transformer.update_transforms([self.transform], transform_time)

    @property
    @abstractmethod
    def pose(self) -> PoseStamped:
        """
        :return: A Pose object containing the pose of the link relative to the world frame.
        """
        ...

    @pose.setter
    @abstractmethod
    def pose(self, pose: PoseStamped) -> None:
        """
        Set the pose of this body.

        :param pose: The pose of this body.
        """
        ...

    @property
    @abstractmethod
    def tf_frame(self) -> str:
        """
        The tf frame of this entity.

        :return: The tf frame of this entity.
        """
        pass

    @property
    def contact_points(self) -> ContactPointsList:
        """
        :return: The contact points of this body with other physical bodies.
        """
        return self.world.get_body_contact_points(self)

    def get_contact_points_with_body(self, body: PhysicalBody) -> ContactPointsList:
        """
        :param body: The body to get the contact points with.
        :return: The contact points of this body with the given body.
        """
        return self.world.get_contact_points_between_two_bodies(self, body)

    def closest_points(self, max_distance: float) -> ClosestPointsList:
        """
        :param max_distance: The maximum distance to consider a body as close, only points closer than or equal to this
         distance will be returned.
        :return: The closest points of this body with other physical bodies within the given maximum distance.
        """
        return self.world.get_body_closest_points(self, max_distance)

    def get_closest_points_with_body(self, body: PhysicalBody, max_distance: float) -> ClosestPointsList:
        """
        :param body: The body to get the points with.
        :param max_distance: The maximum distance to consider a body as close, only points closer than or equal to this
         distance will be returned.
        :return: The closest points of this body with the given body within the given maximum distance.
        """
        return self.world.get_closest_points_between_two_bodies(self, body, max_distance)

    @property
    def is_moving(self) -> Optional[bool]:
        """
        :return: True if this body is moving, False if not, and None if not known.
        """
        if self.is_translating is not None or self.is_rotating is not None:
            return self.is_translating or self.is_rotating
        else:
            return None

    @property
    def is_stationary(self) -> Optional[bool]:
        """
        :return: True if this body is stationary, False otherwise.
        """
        return None if self.is_moving is None else not self.is_moving

    @property
    @abstractmethod
    def color(self) -> Union[Color, Dict[str, Color]]:
        """
        :return: A Color object containing the rgba_color of this body or a dictionary if the body is articulated.
        """
        ...

    @deprecated("Use color property setter instead")
    def set_color(self, color: Color) -> None:
        """
        Set the color of this body, could be rgb or rgba.

        :param color: The color as a list of floats, either rgb or rgba.
        """
        self.color = color

    @color.setter
    @abstractmethod
    def color(self, color: Color) -> None:
        """
        Set the color of this body, could be rgb or rgba.

        :param color: The color as a list of floats, either rgb or rgba.
        """
        ...

    def get_preferred_grasp_alignment(self) -> PreferredGraspAlignment:
        """
        Determines the preferred grasp alignment for an object based on its type.
        This includes the preferred axis, whether grasping from the top is preferred,
        and whether horizontal alignment (90° rotation around X) is preferred.

        :return: The preferred grasp alignment for the object.
        """
        object_type = self.ontology_concept

        try:
            pref_axis = object_type.has_preferred_alignment[0].has_preferred_axis[0].value
            sidegrasp_axis = AxisIdentifier.from_tuple(pref_axis)
            grasp_top = object_type.has_preferred_alignment[0].has_vertical_alignment[0].value
            grasp_horizontal = object_type.has_preferred_alignment[0].has_rotated_gripper[0].value
        except IndexError:
            sidegrasp_axis, grasp_top, grasp_horizontal = (None, False, False)

        preferred_alignment = PreferredGraspAlignment(sidegrasp_axis, grasp_top, grasp_horizontal)

        return preferred_alignment

    def calculate_grasp_descriptions(self, robot: Object) -> List[GraspDescription]:
        """
        Calculates the grasp configurations of an object relative to the robot based on orientation and position.

        This method determines the possible grasp configurations (side and top/bottom faces) of the object,
        taking into account the object's orientation, position, and whether horizontal alignment is preferred.

        :param robot: The robot for which the grasp configurations are being calculated.

        :return: A sorted list of GraspDescription instances representing all grasp permutations.
        """
        objectTmap = self.pose

        preferred_grasp_alignment = self.get_preferred_grasp_alignment()

        grasp_configs = objectTmap.calculate_grasp_descriptions(robot, preferred_grasp_alignment)

        return grasp_configs

    def get_grasp_pose(self, end_effector, grasp: GraspDescription) -> Pose:
        """
        Translates the grasp pose of the object using the desired grasp description and object knowledge.
        Leaves the orientation untouched.
        Returns the translated grasp pose.

        :param end_effector: The end effector that will be used to grasp the object.
        :param grasp: The desired grasp description.

        :return: The grasp pose of the object.
        """
        grasp_pose = self.pose.copy()

        if self.ontology_concept.has_preferred_alignment[0].has_rim_grasp[0].value:
            approach_axis = end_effector.get_approach_axis()
            approach_direction = grasp.approach_direction
            rim_direction = GraspDescription(approach_direction, None, False)
            rim_direction_index = approach_direction.value[0].value.index(1)
            rim_offset = self.get_rotated_bounding_box().dimensions[rim_direction_index] / 2
            grasp_pose.rotate_by_quaternion(end_effector.grasps[rim_direction])
            grasp_pose = LocalTransformer().translate_pose_along_local_axis(grasp_pose, approach_axis, -rim_offset)
            grasp_pose = PoseStamped.from_list(grasp_pose.position.to_list(), self.orientation.to_list())

        return grasp_pose

    def get_approach_offset(self) -> float:
        """
        :return: The pre-grasp offset of the object. It is the largest dimension of the object divided by 2.
        """
        max_object_dimension = max(self.get_rotated_bounding_box().dimensions)
        return max_object_dimension / 2



import time
import unittest
import xml.etree.ElementTree as ET

from pycram.testing import BulletWorldTestCase, BulletWorldGUITestCase
from pycram.datastructures.dataclasses import Color
from pycram.datastructures.enums import WorldMode
from pycram.datastructures.pose import PoseStamped
from pycram.datastructures.world import UseProspectionWorld
from pycram.object_descriptors.urdf import ObjectDescription
from pycram.robot_description import RobotDescription
from pycram.world_concepts.world_object import Object
from pycrap.ontologies import Milk, Robot
from pycram.ros import get_ros_package_path

fix_missing_inertial = ObjectDescription.fix_missing_inertial


class BulletWorldTest(BulletWorldTestCase):

    def test_object_movement(self):
        self.milk.set_position(PoseStamped.from_list([0, 1, 1]))
        self.assertEqual(self.milk.get_position_as_list(), [0, 1, 1])

    def test_robot_orientation(self):
        self.robot.set_pose(PoseStamped.from_list([0, 1, 1]))
        head_position = self.robot.get_link_position('head_pan_link').z
        self.robot.set_orientation(PoseStamped.from_list(orientation=[0, 0, 1, 1]))
        self.assertEqual(self.robot.get_link_position('head_pan_link').z, head_position)

    def test_save_and_restore_state(self):
        self.robot.attach(self.milk)
        self.milk.attach(self.cereal)
        all_object_attachments = {obj: obj.attachments.copy() for obj in self.world.objects}
        state_id = self.world.save_state()
        self.milk.detach(self.cereal)
        robot_link = self.robot.root_link
        milk_link = self.milk.root_link
        cid = robot_link.constraint_ids[milk_link]
        self.assertTrue(cid == self.robot.attachments[self.milk].id)
        self.world.remove_constraint(cid)
        self.world.restore_state(state_id)
        cid = robot_link.constraint_ids[milk_link]
        self.assertTrue(milk_link in robot_link.constraint_ids)
        self.assertTrue(cid == self.robot.attachments[self.milk].id)
        for obj in self.world.objects:
            self.assertTrue(len(obj.attachments) == len(all_object_attachments[obj]))
            for att in obj.attachments:
                self.assertTrue(att in all_object_attachments[obj])

    def test_remove_object(self):
        milk_id = self.milk.id
        self.assertTrue(milk_id in [obj.id for obj in self.world.objects])
        self.world.remove_object(self.milk)
        self.assertTrue(milk_id not in [obj.id for obj in self.world.objects])
        BulletWorldTest.milk = Object("milk", Milk, "milk.stl", pose=PoseStamped.from_list([1.3, 1, 0.9]))

    def test_remove_robot(self):
        robot_id = self.robot.id
        self.assertTrue(robot_id in [obj.id for obj in self.world.objects])
        self.world.remove_object(self.robot)
        self.assertTrue(robot_id not in [obj.id for obj in self.world.objects])
        BulletWorldTest.robot = Object(RobotDescription.current_robot_description.name, Robot,
                                       RobotDescription.current_robot_description.name + self.extension)

    def test_get_joint_position(self):
        self.assertAlmostEqual(self.robot.get_joint_position("head_pan_joint"), 0.0, delta=0.01)

    def test_get_object_contact_points(self):
        self.assertEqual(len(self.robot.contact_points), 0)  # 8 because of robot wheels with floor
        self.milk.set_position(self.robot.get_position())
        self.assertTrue(len(self.robot.contact_points) > 0)

    def test_enable_joint_force_torque_sensor(self):
        self.world.enable_joint_force_torque_sensor(self.robot, self.robot.get_joint_id("head_pan_joint"))
        force_torque = self.world.get_joint_reaction_force_torque(self.robot, self.robot.get_joint_id("head_pan_joint"))
        # TODO: useless because even if the sensor is disabled, the force_torque is still 0
        for ft in force_torque:
            self.assertTrue(ft == 0.0)

    def test_disable_joint_force_torque_sensor(self):
        self.world.enable_joint_force_torque_sensor(self.robot, self.robot.get_joint_id("head_pan_joint"))
        self.world.disable_joint_force_torque_sensor(self.robot, self.robot.get_joint_id("head_pan_joint"))
        force_torque = self.world.get_joint_reaction_force_torque(self.robot, self.robot.get_joint_id("head_pan_joint"))
        for ft in force_torque:
            self.assertTrue(ft == 0.0)

    def test_get_applied_joint_motor_torque(self):
        self.world.get_applied_joint_motor_torque(self.robot, self.robot.get_joint_id("head_pan_joint"))

    def test_step_simulation(self):
        # TODO: kitchen explodes when stepping simulation, fix this
        self.kitchen.set_position([100, 100, 0])
        self.milk.set_position(PoseStamped.from_list([0, 0, 2]))
        self.world.simulate(1)
        self.assertTrue(self.milk.get_position().z < 2)

    @unittest.skip
    def test_set_real_time_simulation(self):
        self.milk.set_position(PoseStamped.from_list([100, 0, 2]))
        curr_time = time.time()
        self.world.simulate(0.5, real_time=True)
        time_elapsed = time.time() - curr_time
        self.assertAlmostEqual(time_elapsed, 0.5, delta=0.2)

    def test_collision_callback(self):
        self.kitchen.set_position([100, 100, 0])
        self.collision_called = False
        self.no_collision_called = False

        def collision_callback():
            self.collision_called = True

        def no_collision_callback():
            self.no_collision_called = True

        self.world.register_two_objects_collision_callbacks(self.milk, self.cereal,
                                                            collision_callback, no_collision_callback)

        self.world.simulate(1)
        self.assertTrue(self.no_collision_called)
        self.assertFalse(self.collision_called)

        self.collision_called = False
        self.no_collision_called = False

        new_milk_position = self.cereal.get_position()
        new_milk_position.z += 0.5
        self.milk.set_position(new_milk_position)

        self.world.simulate(4)
        self.assertTrue(self.collision_called)

    def test_equal_world_states(self):
        time.sleep(2.5)
        self.robot.set_pose(PoseStamped.from_list([1, 0, 0], [0, 0, 0, 1]))
        self.assertFalse(self.world.world_sync.check_for_equal())
        with UseProspectionWorld():
            self.assertTrue(self.world.world_sync.check_for_equal())

    def test_add_resource_path(self):
        self.world.add_resource_path("test")
        self.assertTrue("test" in self.world.get_data_directories())

    def test_no_prospection_object_found_for_given_object(self):
        milk_2 = Object("milk_2", Milk, "milk.stl", pose=PoseStamped.from_list([1.3, 1, 0.9]))
        try:
            prospection_milk_2 = self.world.get_prospection_object_for_object(milk_2)
            self.world.remove_object(milk_2)
            self.world.get_prospection_object_for_object(milk_2)
            self.assertFalse(True)
        except KeyError as e:
            self.assertTrue(True)

    def test_real_object_position_does_not_change_with_prospection_object(self):
        milk_2_pos = [1.3, 1, 0.9]
        milk_2 = Object("milk_3", Milk, "milk.stl", pose=PoseStamped.from_list(milk_2_pos))
        time.sleep(0.05)
        milk_2_pos = milk_2.get_position()

        # Assert that when prospection object is moved, the real object is not moved
        with UseProspectionWorld():
            prospection_milk = self.world.get_prospection_object_for_object(milk_2)
            prospection_milk_pos = prospection_milk.get_position()
            self.assertTrue(prospection_milk_pos == milk_2_pos)
            prospection_milk_pos.x += 1
            prospection_milk.set_position(prospection_milk_pos)
            self.assertTrue(prospection_milk.get_position() != milk_2.get_position())
        self.world.remove_object(milk_2)

    def test_prospection_object_position_does_not_change_with_real_object(self):
        milk_2_pos = [1.3, 1, 0.9]
        milk_2 = Object("milk_4", Milk, "milk.stl", pose=PoseStamped.from_list(milk_2_pos))
        time.sleep(0.05)
        milk_2_pos = milk_2.get_position()

        # Assert that when real object is moved, the prospection object is not moved
        with UseProspectionWorld():
            prospection_milk = self.world.get_prospection_object_for_object(milk_2)
            prospection_milk_pos = prospection_milk.get_position()
            self.assertTrue(prospection_milk_pos == milk_2_pos)
            milk_2_pos.x += 1
            milk_2.set_position(milk_2_pos)
            self.assertTrue(prospection_milk.get_position() != milk_2.get_position())
        self.world.remove_object(milk_2)

    def test_add_vis_axis(self):
        self.world.add_vis_axis(self.robot.get_link_pose(RobotDescription.current_robot_description.get_camera_link()))
        self.assertTrue(len(self.world.vis_axis) == 1)
        self.world.remove_vis_axis()
        self.assertTrue(len(self.world.vis_axis) == 0)

    def test_add_text(self):
        link: ObjectDescription.Link = self.robot.get_link(RobotDescription.current_robot_description.get_camera_link())
        text_id = self.world.add_text("test", link.position.to_list(), link.orientation.to_list(), 1,
                                      Color(1, 0, 0, 1), 3, link.object_id, link.id)
        if self.world.mode == WorldMode.GUI:
            time.sleep(4)

    def test_remove_text(self):
        link: ObjectDescription.Link = self.robot.get_link(RobotDescription.current_robot_description.get_camera_link())
        text_id_1 = self.world.add_text("test 1", link.pose.position.to_list(), link.pose.orientation.to_list(), 1,
                                        Color(1, 0, 0, 1), 0, link.object_id, link.id)
        text_id = self.world.add_text("test 2", link.pose.position.to_list(), link.pose.orientation.to_list(), 1,
                                      Color(0, 1, 0, 1), 0, link.object_id, link.id)

        if self.world.mode == WorldMode.GUI:
            time.sleep(2)
        self.world.remove_text(text_id_1)
        if self.world.mode == WorldMode.GUI:
            time.sleep(3)

    def test_remove_all_text(self):
        link: ObjectDescription.Link = self.robot.get_link(RobotDescription.current_robot_description.get_camera_link())
        text_id_1 = self.world.add_text("test 1", link.pose.position.to_list(), link.pose.orientation.to_list(), 1,
                                        Color(1, 0, 0, 1), 0, link.object_id, link.id)
        text_id = self.world.add_text("test 2", link.pose.position.to_list(), link.pose.orientation.to_list(), 1,
                                      Color(0, 1, 0, 1), 0, link.object_id, link.id)
        if self.world.mode == WorldMode.GUI:
            time.sleep(2)
        self.world.remove_text()
        if self.world.mode == WorldMode.GUI:
            time.sleep(3)


@unittest.skip("Not working in CI")
class BulletWorldTestGUI(BulletWorldGUITestCase):
    def test_add_vis_axis(self):
        time.sleep(10)
        self.world.add_vis_axis(self.robot.get_link_pose(RobotDescription.current_robot_description.get_camera_link()))
        self.assertTrue(len(self.world.vis_axis) == 1)
        self.world.remove_vis_axis()
        self.assertTrue(len(self.world.vis_axis) == 0)


class XMLTester(unittest.TestCase):

    def setUp(self) -> None:
        filename = get_ros_package_path('pycram') + '/resources/robots/' + 'pr2.urdf'
        with open(filename, "r") as f:
            self.urdf_string = f.read()

    def test_inertial(self):
        result = fix_missing_inertial(self.urdf_string)
        resulting_tree = ET.ElementTree(ET.fromstring(result))
        for element in resulting_tree.iter("link"):
            self.assertTrue(len([*element.iter("inertial")]) > 0)



# used for delayed evaluation of typing until python 3.11 becomes mainstream
from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, fields
from datetime import timedelta

from pycrap.ontologies import PhysicalObject, Agent
from .datastructures.enums import ObjectType
from .datastructures.pose import PoseStamped, GraspDescription
from .failures import PlanFailure

from .datastructures.world import World
from .datastructures.partial_designator import PartialDesignator
from typing_extensions import Type, List, Dict, Any, Optional, Union, Callable, Iterable, TYPE_CHECKING, ForwardRef, \
    Self, Iterator
from typing import get_type_hints

from .has_parameters import HasParameters
from .language import LanguageMixin
from .local_transformer import LocalTransformer
from .robot_description import RobotDescription
from .ros import loginfo
from .utils import bcolors, is_iterable
from .world_concepts.world_object import Object as WorldObject, Object
from .plan import Plan


class DesignatorError(Exception):
    """Implementation of designator_description errors."""

    def __init__(self, *args, **kwargs):
        """Create a new designator_description error."""
        Exception.__init__(self, *args, **kwargs)


class ResolutionError(Exception):
    def __init__(self, missing_properties: List[str], wrong_type: Dict, current_type: Any,
                 designator: DesignatorDescription):
        self.error = (f"\nSome required properties where missing or had the wrong type when grounding the Designator:"
                      f" {designator}.\n")
        self.missing = f"The missing properties where: {missing_properties}\n"
        self.wrong = f"The properties with the wrong type along with the current -and right type :\n"
        self.head = ("Property   |   Current Type    |     Right Type\n------------------------------------"
                     "-------------------------\n")
        self.tab = ""
        for prop in wrong_type.keys():
            self.tab += prop + "     " + str(current_type[prop]) + "      " + str(wrong_type[prop]) + "\n"
        self.message = self.error
        if missing_properties != []:
            self.message += self.missing
        if wrong_type != {}:
            self.message += self.wrong + self.head + self.tab
        self.message = f"{bcolors.BOLD}{bcolors.FAIL}" + self.message + f"{bcolors.ENDC}"
        super(ResolutionError, self).__init__(self.message)


class DesignatorDescription:
    """
    :ivar resolve: The specialized_designators function to use for this designator_description, defaults to self.ground
    """

    def __init__(self):
        """
        Create a Designator description.
        """
        pass

    def resolve(self):
        return self.ground()

    def make_dictionary(self, properties: List[str]):
        """
        Creates a dictionary of this description with only the given properties
        included.

        :param properties: A list of properties that should be included in the dictionary.
                            The given properties have to be an attribute of this description.
        :return: A dictionary with the properties as keys.
        """
        attributes = self.__dict__
        ret = {}
        for att in attributes.keys():
            if att in properties:
                ret[att] = attributes[att]
        return ret

    def ground(self) -> Any:
        """
        Should be overwritten with an actual grounding function which infers missing properties.
        """
        return self

    def get_slots(self) -> List[str]:
        """
        :return: a list of all slots of this description. Can be used for inspecting different descriptions and
         debugging.
        """
        return list(self.__dict__.keys())

    def copy(self) -> DesignatorDescription:
        return self

    def get_optional_parameter(self) -> List[str]:
        """
        Returns a list of optional parameter names of this designator_description description.
        """
        return [param_name for param_name, param in inspect.signature(self.__init__).parameters.items() if
                param.default != param.empty]

    def get_all_parameter(self) -> List[str]:
        """
        Returns a list of all parameter names of this designator_description description.
        """
        return [param_name for param_name, param in inspect.signature(self.__init__).parameters.items()]
    @classmethod
    def get_type_hints(cls) -> Dict[str, Any]:
        """
        Returns the type hints of the __init__ method of this designator_description description.

        :return:
        """
        return get_type_hints(cls.__init__)

@dataclass
class ActionDescription(HasParameters):
    """
    The performable designator_description with a single element for each list of possible parameter.
    """
    robot_position: PoseStamped = field(init=False)
    """
    The position of the robot at the start of the action.
    """
    robot_torso_height: float = field(init=False)
    """
    The torso height of the robot at the start of the action.
    """

    _robot_type: Type[Agent] = field(init=False)
    """
    The type of the robot at the start of the action.
    """
    _pre_perform_callbacks = []
    """
    List of callback functions that will be called before the action is performed.
    """

    _post_perform_callbacks = []
    """
    List of callback functions that will be called after the action is performed.
    """

    def __post_init__(self):
      self._pre_perform_callbacks.append(self._update_robot_params)

    def perform(self) -> Any:
        """
        Executes the action with the single parameters from the description.

        :return: The result of the action in the plan
        """
        result: Optional[Any] = None
        for pre_perform in self._pre_perform_callbacks:
            pre_perform(self)
        try:
            result = self.plan()
        except PlanFailure as e:
            raise e
        finally:
            for post_perform in self._post_perform_callbacks:
                post_perform(self)
            self.validate(result)
        return result

    def plan(self) -> Any:
        """
        Plan of the action. To be overridden by subclasses.

        :return: The result of the action, if there is any
        """
        raise NotImplementedError()

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        """
        Validate the action after performing it, by checking if the action effects are as expected.

        :param result: The result of the action if there is any
        :param max_wait_time: The maximum time to wait for the action to be validated, before raising an error.
        """
        raise NotImplementedError()

    @classmethod
    def get_type_hints(cls, localns=None) -> Dict[str, Any]:
        """
        Returns the type hints of the __init__ method of this designator_description description.

        :return:
        """
        l = localns if localns is not None else locals()
        return get_type_hints(cls, localns=l)

    @classmethod
    def pre_perform(cls, func) -> Callable:
        """
        Decorator to execute the decorated function before performing the action.

        :param func: The function to be decorated.
        :return: The decorated function.
        """
        cls._pre_perform_callbacks.append(func)

        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    @classmethod
    def post_perform(cls, func) -> Callable:
        """
        Decorator to execute the decorated function after performing the action.

        :param func: The function to be decorated.
        :return: The decorated function.
        """
        cls._post_perform_callbacks.append(func)

        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    @classmethod
    def description(cls, *args, **kwargs) -> PartialDesignator[Self]:
        raise NotImplementedError()

    def __str__(self):
        # all fields that are not ORM classes
        fields = {}
        for key, value in vars(self).items():
            if key.startswith("orm_"):
                continue
            if isinstance(value, Object):
                fields[key] = value.name
            elif isinstance(value, PoseStamped):
                fields[key] = value.__str__()
        fields_str = "\n".join([f"{key}: {value}" for key, value in fields.items()])
        return f"{self.__class__.__name__.replace('Performable', '')}:\n{fields_str}"

    def __repr__(self):
        return self.__str__()

    def _update_robot_params(self, action: ActionDescription):
        action.robot_position = World.robot.pose
        action.robot_torso_height = World.robot.get_joint_position(RobotDescription.current_robot_description.torso_joint)
        action._robot_type = World.robot.obj_type

class LocationDesignatorDescription(DesignatorDescription, PartialDesignator):
    """
    Parent class of location designator_description descriptions.
    """

    def __init__(self):
        super().__init__()

    def ground(self) -> PoseStamped:
        """
        Find a location that satisfies all constrains.
        """
        raise NotImplementedError(f"{type(self)}.ground() is not implemented.")


class ObjectDesignatorDescription(DesignatorDescription, PartialDesignator):
    """
    Class for object designator_description descriptions.
    Descriptions hold possible parameter ranges for object designators.
    """
    def __init__(self, names: Optional[List[str]] = None, types: Optional[List[Type[PhysicalObject]]] = None):
        """
        Base of all object designator_description descriptions. Every object designator_description has the name and type of the object.

        :param names: A list of names that could describe the object
        :param types: A list of types that could represent the object
        """
        super().__init__()
        PartialDesignator.__init__(self, ObjectDesignatorDescription, names=names, types=types)
        self.types: Optional[List[ObjectType]] = types
        self.names: Optional[List[str]] = names

    def ground(self) -> WorldObject:
        """
        Return the first object from the world that fits the description.

        :return: A executed object designator_description
        """
        return next(iter(self))

    def __iter__(self) -> Iterator[WorldObject]:
        """
        Iterate through all possible objects fitting this description

        :yield: A executed object designator_description
        """
        for params in self.generate_permutations():

            # for every world object
            for obj in World.current_world.objects:

                # skip if name does not match specification
                if self.names and obj.name not in params.values():
                    continue

                # skip if type does not match specification
                if self.types and obj.obj_type not in params.values():
                    continue

                # yield self.Object(obj.name, obj.obj_type, obj)
                yield obj

    def flatten(self) -> List:
        res = [None] * 7
        res.append(self.types[0])
        return res


@dataclass
class BaseMotion(ABC):

    @abstractmethod
    def perform(self):
        """
        Passes this designator to the process module for execution. Will be overwritten by each motion.
        """
        pass

    def __post_init__(self):
        """
        Checks if types are missing or wrong
        """
        return
        # TODO include type checks for this again (use type guard?)
        #
        # right_types = get_type_hints(self)
        # attributes = self.__dict__.copy()
        #
        # missing = []
        # wrong_type = {}
        # current_type = {}
        #
        # for k in attributes.keys():
        #     attribute = attributes[k]
        #     attribute_type = type(attributes[k])
        #     right_type = right_types[k]
        #     types = get_args(right_type)
        #     if attribute is None:
        #         if not any([x is type(None) for x in get_args(right_type)]):
        #             missing.append(k)
        #     elif not issubclass(attribute_type, right_type): # attribute_type is not right_type:
        #         if attribute_type not in types:
        #             if attribute_type not in [get_origin(x) for x in types if x is not type(None)]:
        #                 wrong_type[k] = right_types[k]
        #                 current_type[k] = attribute_type
        # if missing != [] or wrong_type != {}:
        #     raise ResolutionError(missing, wrong_type, current_type, self.__class__)
        #



"""Module holding all enums of PyCRAM."""

from enum import Enum, auto, IntEnum

from ..failures import UnsupportedJointType


class AdjacentBodyMethod(Enum):
    ClosestPoints = auto()
    """
    The ClosestPoints method is used to find the closest points in other bodies to the body.
    """
    RayCasting = auto()
    """
    The RayCasting method is used to find the points in other bodies that are intersected by rays cast
     from the body bounding box to 6 directions (up, down, left, right, front, back).
    """


class ContainerManipulationType(Enum):
    """
    Enum for the different types of container manipulation.
    """
    Opening = auto()
    """
    The Opening type is used to open a container.
    """
    Closing = auto()
    """
    The Closing type is used to close a container.
    """


class FindBodyInRegionMethod(Enum):
    """
    Enum for the different methods to find a body in a region.
    """
    FingerToCentroid = auto()
    """
    The FingerToCentroid method is used to find the body in a region by casting a ray from each finger to the
     centroid of the region.
    """
    Centroid = auto()
    """
    The Centroid method is used to find the body in a region by calculating the centroid of the region and
    casting two rays from opposite sides of the region to the centroid.
    """
    MultiRay = auto()
    """
    The MultiRay method is used to find the body in a region by casting multiple rays covering the region.
    """


class Frame(Enum):
    Map = "map"


class StaticJointState(Enum):
    Park = "park"


class DescriptionType(Enum):
    URDF = "urdf"
    MJCF = "mjcf"

    def get_file_extension(self):
        if self == DescriptionType.URDF:
            return ".urdf"
        elif self == DescriptionType.MJCF:
            return ".xml"
        else:
            raise ValueError("Unknown description type")


class ExecutionType(Enum):
    """Enum for Execution Process Module types."""
    REAL = auto()
    SIMULATED = auto()
    SEMI_REAL = auto()


class Arms(int, Enum):
    """Enum for Arms."""
    LEFT = 0
    RIGHT = 1
    BOTH = 2

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


class TaskStatus(int, Enum):
    """
    Enum for readable descriptions of a tasks' status.
    """
    CREATED = 0
    RUNNING = 1
    SUCCEEDED = 2
    FAILED = 3
    INTERRUPTED = 4
    SLEEPING = 5


class JointType(Enum):
    """
    Enum for readable joint types.
    """
    REVOLUTE = 0
    PRISMATIC = 1
    SPHERICAL = 2
    PLANAR = 3
    FIXED = 4
    UNKNOWN = 5
    CONTINUOUS = 6
    FLOATING = 7


class AxisIdentifier(Enum):
    """
    Enum for translating the axis name to a vector along that axis.
    """
    X = (1, 0, 0)
    Y = (0, 1, 0)
    Z = (0, 0, 1)

    @classmethod
    def from_tuple(cls, axis_tuple):
        return next((axis for axis in cls if axis.value == axis_tuple), None)


class Grasp(Enum):
    """
    Enum for Grasp orientations.
    """
    FRONT = (AxisIdentifier.X, -1)
    BACK = (AxisIdentifier.X, 1)
    RIGHT = (AxisIdentifier.Y, -1)
    LEFT = (AxisIdentifier.Y, 1)
    TOP = (AxisIdentifier.Z, -1)
    BOTTOM = (AxisIdentifier.Z, 1)

    @classmethod
    def from_axis_direction(cls, axis: AxisIdentifier, direction: int):
        """Get the Grasp face from an axis-index tuple"""
        return next((grasp for grasp in cls if grasp.value == (axis, direction)), None)


class ObjectType(int, Enum):
    """
    Enum for Object types to easier identify different objects
    """
    METALMUG = auto()
    PRINGLES = auto()
    MILK = auto()
    SPOON = auto()
    BOWL = auto()
    BREAKFAST_CEREAL = auto()
    JEROEN_CUP = auto()
    ROBOT = auto()
    GRIPPER = auto()
    ENVIRONMENT = auto()
    GENERIC_OBJECT = auto()
    HUMAN = auto()
    IMAGINED_SURFACE = auto()


class Shape(Enum):
    """
    Enum for visual shapes of objects
    """
    SPHERE = 2
    BOX = 3
    CYLINDER = 4
    MESH = 5
    PLANE = 6
    CAPSULE = 7


class TorsoState(IntEnum):
    """
    Enum for the different states of the torso.
    """
    HIGH = auto()
    MID = auto()
    LOW = auto()


class WorldMode(Enum):
    """
    Enum for the different modes of the world.
    """
    GUI = "GUI"
    DIRECT = "DIRECT"


class GripperState(Enum):
    """
    Enum for the different motions of the gripper.
    """
    OPEN = auto()
    CLOSE = auto()

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


class GripperType(Enum):
    """
    Enum for the different types of grippers.
    """
    PARALLEL = auto()
    SUCTION = auto()
    FINGER = auto()
    HYDRAULIC = auto()
    PNEUMATIC = auto()
    CUSTOM = auto()


class ImageEnum(Enum):
    """
    Enum for image switch view on hsrb display.
    """
    HI = 0
    TALK = 1
    DISH = 2
    DONE = 3
    DROP = 4
    HANDOVER = 5
    ORDER = 6
    PICKING = 7
    PLACING = 8
    REPEAT = 9
    SEARCH = 10
    WAVING = 11
    FOLLOWING = 12
    DRIVINGBACK = 13
    PUSHBUTTONS = 14
    FOLLOWSTOP = 15
    JREPEAT = 16
    SOFA = 17
    INSPECT = 18
    CHAIR = 37


class DetectionTechnique(int, Enum):
    """
    Enum for techniques for detection tasks.
    """
    ALL = 0
    HUMAN = 1
    TYPES = 2
    REGION = 3
    HUMAN_ATTRIBUTES = 4
    HUMAN_WAVING = 5


class DetectionState(int, Enum):
    """
    Enum for the state of the detection task.
    """
    START = 0
    STOP = 1
    PAUSE = 2


class LoggerLevel(Enum):
    """
    Enum for the different logger levels.
    """
    DEBUG = 'debug'
    INFO = 'info'
    WARN = 'warn'
    ERROR = 'error'
    FATAL = 'fatal'


class VirtualMobileBaseJointName(Enum):
    """
    Enum for the joint names of the virtual mobile base.
    """
    LINEAR_X = "odom_vel_lin_x_joint"
    LINEAR_Y = "odom_vel_lin_y_joint"
    ANGULAR_Z = "odom_vel_ang_z_joint"


class MJCFGeomType(Enum):
    """
    Enum for the different geom types in a MuJoCo XML file.
    """
    BOX = "box"
    CYLINDER = "cylinder"
    CAPSULE = "capsule"
    SPHERE = "sphere"
    PLANE = "plane"
    MESH = "mesh"
    ELLIPSOID = "ellipsoid"
    HFIELD = "hfield"
    SDF = "sdf"


MJCFBodyType = MJCFGeomType
"""
Alias for MJCFGeomType. As the body type is the same as the geom type.
"""


class MJCFJointType(Enum):
    """
    Enum for the different joint types in a MuJoCo XML file.
    """
    FREE = "free"
    BALL = "ball"
    SLIDE = "slide"
    HINGE = "hinge"
    FIXED = "fixed"  # Added for compatibility with PyCRAM, but not a real joint type in MuJoCo.


class MovementType(Enum):
    """
    Enum for the different movement types of the robot.
    """
    STRAIGHT_TRANSLATION = auto()
    STRAIGHT_CARTESIAN = auto()
    TRANSLATION = auto()
    CARTESIAN = auto()

class WaypointsMovementType(Enum):
    """
    Enum for the different movement types of the robot.
    """
    ENFORCE_ORIENTATION_STRICT = auto()
    ENFORCE_ORIENTATION_FINAL_POINT = auto()


class MultiverseAPIName(Enum):
    """
    Enum for the different APIs of the Multiverse.
    """
    GET_CONTACT_POINTS = "get_contact_points"
    GET_CONTACT_BODIES = "get_contact_bodies"
    GET_CONTACT_BODIES_AND_POINTS = "get_contact_bodies_and_points"
    GET_CONSTRAINT_EFFORT = "get_constraint_effort"
    GET_BOUNDING_BOX = "get_bounding_box"
    ATTACH = "attach"
    DETACH = "detach"
    GET_RAYS = "get_rays"
    EXIST = "exist"
    PAUSE = "pause"
    UNPAUSE = "unpause"
    SAVE = "save"
    LOAD = "load"


class MultiverseProperty(Enum):
    def __str__(self):
        return self.value


class MultiverseBodyProperty(MultiverseProperty):
    """
    Enum for the different properties of a body the Multiverse.
    """
    POSITION = "position"
    ORIENTATION = "quaternion"
    RELATIVE_VELOCITY = "relative_velocity"


class MultiverseJointProperty(MultiverseProperty):
    pass


class MultiverseJointPosition(MultiverseJointProperty):
    """
    Enum for the Position names of the different joint types in the Multiverse.
    """
    REVOLUTE_JOINT_POSITION = "joint_rvalue"
    PRISMATIC_JOINT_POSITION = "joint_tvalue"

    @classmethod
    def from_pycram_joint_type(cls, joint_type: JointType) -> 'MultiverseJointPosition':
        if joint_type in [JointType.REVOLUTE, JointType.CONTINUOUS]:
            return MultiverseJointPosition.REVOLUTE_JOINT_POSITION
        elif joint_type == JointType.PRISMATIC:
            return MultiverseJointPosition.PRISMATIC_JOINT_POSITION
        else:
            raise UnsupportedJointType(joint_type)


class MultiverseJointCMD(MultiverseJointProperty):
    """
    Enum for the Command names of the different joint types in the Multiverse.
    """
    REVOLUTE_JOINT_CMD = "cmd_joint_rvalue"
    PRISMATIC_JOINT_CMD = "cmd_joint_tvalue"

    @classmethod
    def from_pycram_joint_type(cls, joint_type: JointType) -> 'MultiverseJointCMD':
        if joint_type in [JointType.REVOLUTE, JointType.CONTINUOUS]:
            return MultiverseJointCMD.REVOLUTE_JOINT_CMD
        elif joint_type == JointType.PRISMATIC:
            return MultiverseJointCMD.PRISMATIC_JOINT_CMD
        else:
            raise UnsupportedJointType(joint_type)


class FilterConfig(Enum):
    """
    Declare existing filter methods.
    Currently supported: Butterworth
    """
    butterworth = 1



from typing_extensions import Optional, Type, Union, Dict

from ...worlds.multiverse_communication.clients import MultiverseWriter, MultiverseAPI, MultiverseClient, \
    MultiverseReader, MultiverseController

from ...config.multiverse_conf import MultiverseConfig as Conf


class MultiverseClientManager:
    BASE_PORT: int = Conf.BASE_CLIENT_PORT
    """
    The base port of the Multiverse client.
    """
    clients: Optional[Dict[str, MultiverseClient]] = {}
    """
    The list of Multiverse clients.
    """
    last_used_port: int = BASE_PORT

    def __init__(self, simulation_wait_time_factor: Optional[float] = 1.0):
        """
        Initialize the Multiverse client manager.

        :param simulation_wait_time_factor: The simulation wait time factor.
        """
        self.simulation_wait_time_factor = simulation_wait_time_factor

    def create_reader(self, is_prospection_world: Optional[bool] = False) -> MultiverseReader:
        """
        Create a Multiverse reader client.

        :param is_prospection_world: Whether the reader is connected to the prospection world.
        """
        return self.create_client(MultiverseReader, "reader", is_prospection_world)

    def create_writer(self, simulation: str, is_prospection_world: Optional[bool] = False) -> MultiverseWriter:
        """
        Create a Multiverse writer client.

        :param simulation: The name of the simulation that the writer is connected to
         (usually the name defined in the .muv file).
        :param is_prospection_world: Whether the writer is connected to the prospection world.
        """
        return self.create_client(MultiverseWriter, "writer", is_prospection_world,
                                  simulation=simulation)

    def create_controller(self, is_prospection_world: Optional[bool] = False) -> MultiverseController:
        """
        Create a Multiverse controller client.

        :param is_prospection_world: Whether the controller is connected to the prospection world.
        """
        return self.create_client(MultiverseController, "controller", is_prospection_world)

    def create_api_requester(self, simulation: str, is_prospection_world: Optional[bool] = False) -> MultiverseAPI:
        """
        Create a Multiverse API client.

        :param simulation: The name of the simulation that the API is connected to
         (usually the name defined in the .muv file).
        :param is_prospection_world: Whether the API is connected to the prospection world.
        """
        return self.create_client(MultiverseAPI, "api_requester", is_prospection_world, simulation=simulation)

    def create_client(self,
                      client_type: Type[MultiverseClient],
                      name: Optional[str] = None,
                      is_prospection_world: Optional[bool] = False,
                      **kwargs) -> Union[MultiverseClient, MultiverseAPI,
                                         MultiverseReader, MultiverseWriter, MultiverseController]:
        """
        Create a Multiverse client.

        :param client_type: The type of the client to create.
        :param name: The name of the client.
        :param is_prospection_world: Whether the client is connected to the prospection world.
        :param kwargs: Any other keyword arguments that should be passed to the client constructor.
        """
        MultiverseClientManager.last_used_port += 1
        name = (name or client_type.__name__) + f"_{self.last_used_port}"
        client = client_type(name, self.last_used_port, is_prospection_world=is_prospection_world,
                             simulation_wait_time_factor=self.simulation_wait_time_factor, **kwargs)
        self.clients[name] = client
        return client

    @classmethod
    def stop_all_clients(cls):
        """
        Stop all clients.
        """
        for client in cls.clients:
            if isinstance(client, MultiverseReader):
                client.stop_thread = True
                client.join()
            elif isinstance(client, MultiverseClient):
                client.stop()
        cls.clients = {}



from owlready2 import Thing, ThingClass, ObjectProperty, FunctionalProperty, get_ontology, And, Or, Not, OneOf, Inverse, normstr, DatatypeProperty, TransitiveProperty, SymmetricProperty, AsymmetricProperty, ReflexiveProperty, IrreflexiveProperty, datetime
import tempfile


ontology_file = tempfile.NamedTemporaryFile()
ontology = get_ontology("file://" + ontology_file.name).load()
CRAX_ONTOLOGY_NAME = "PyCRAP"


class Base(Thing, metaclass=ThingClass):
    namespace = ontology


class BaseProperty(ObjectProperty):
    namespace = ontology


class BaseDatatype(DatatypeProperty):
    namespace = ontology



import tempfile

import owlready2
from typing_extensions import Dict, Any

from .ontologies.base import Base, ontology as default_pycrap_ontology, ontology_file as default_pycrap_ontology_file, \
    CRAX_ONTOLOGY_NAME


class OntologyWrapper:
    """
    Wrapper class for user-friendly access of the owlready2 ontology class.

    This class spawns a temporary file that is used to store the ontology.
    This has to be done whenever several PyCRAP ontologies are needed that store different individuals.
    """

    ontology: owlready2.Ontology
    """
    The owlready2 ontology that is used for reasoning.
    """

    file: tempfile.NamedTemporaryFile
    """
    The file that the ontology is stored in.
    """

    python_objects: Dict[Base, Any]
    """
    A dictionary that maps ontology individuals to python objects.
    """

    def __init__(self):
        self.file = default_pycrap_ontology_file
        self.ontology = default_pycrap_ontology
        self.ontology.name = CRAX_ONTOLOGY_NAME
        self.python_objects = {}

    @property
    def path(self) -> str:
        return self.file.name

    def individuals(self):
        return self.ontology.individuals()

    def destroy_individuals(self):
        """
        Destroys all individuals in the ontology.
        """
        for individual in self.individuals():
            owlready2.destroy_entity(individual)

    @staticmethod
    def classes():
        """
        :return: All classes of the PyCRAP ontology.
        """
        return default_pycrap_ontology.classes()

    def search(self, *args, **kwargs):
        """
        Check https://owlready2.readthedocs.io/en/latest/onto.html#simple-queries for details.

        :return: The search results of the ontology.
        """
        return self.ontology.search(*args, **kwargs)

    def search_one(self, *args, **kwargs):
        return self.ontology.search(*args, **kwargs)

    def __enter__(self):
        return self.ontology.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.ontology.__exit__(exc_type, exc_val, exc_tb)

    def reason(self):
        """
        Reason over the ontology. This may take a long time.
        """
        owlready2.sync_reasoner(self.ontology, infer_property_values=True, debug=False)

    def add_individual(self, individual: Base, python_object: Any):
        self.python_objects[individual] = python_object



import time
import unittest
from datetime import datetime

from random_events.interval import closed
from random_events.product_algebra import SimpleEvent, Event

from pycram.datastructures.enums import TaskStatus
from pycram.designators.action_designator import *
from pycram.language import SequentialPlan, ParallelPlan, CodeNode
from pycram.parameterizer import Parameterizer
from pycram.plan import PlanNode, Plan
from pycram.process_module import simulated_robot
from pycram.testing import BulletWorldTestCase


class TestPlan(unittest.TestCase):

    def setUp(self):
        Plan.current_plan = None

    def test_plan_construction(self):
        node = PlanNode()
        plan = Plan(node)
        self.assertEqual(node, plan.root)
        self.assertEqual(len(plan.edges), 0)
        self.assertEqual(len(plan.nodes), 1)
        self.assertEqual(plan, node.plan)
        self.assertIsNone(Plan.current_plan)

    def test_add_edge(self):
        node = PlanNode()
        plan = Plan(node)
        node2 = PlanNode()
        plan.add_edge(node, node2)
        self.assertEqual(node, plan.root)
        self.assertIn(node, plan.nodes)
        self.assertEqual(len(plan.nodes), 2)
        self.assertEqual(len(plan.edges), 1)
        self.assertIn(node2, plan.nodes)
        self.assertEqual(plan, node2.plan)
        self.assertEqual(plan, node2.plan)

    def test_add_node(self):
        node = PlanNode()
        plan = Plan(node)
        node2 = PlanNode()
        plan.add_node(node2)
        self.assertEqual(node, plan.root)
        self.assertIn(node, plan.nodes)
        self.assertIn(node2, plan.nodes)
        self.assertNotIn((node, node2), plan.edges)
        self.assertEqual(plan, node2.plan)

    def test_mount(self):
        plan1_node = PlanNode()
        plan1 = Plan(plan1_node)
        plan2_node = PlanNode()
        plan2 = Plan(plan2_node)

        plan1.mount(plan2)
        self.assertIn(plan2_node, plan1.nodes)
        self.assertEqual(plan1, plan2_node.plan)
        self.assertEqual(plan1, plan2_node.plan)
        self.assertEqual(len(plan1.edges), 1)
        self.assertEqual(len(plan1.nodes), 2)

    def test_mount_specific_node(self):
        plan = Plan(PlanNode())
        mount_node = PlanNode()
        plan.add_edge(plan.root, mount_node)

        plan2 = Plan(PlanNode())
        plan.mount(plan2, mount_node)

        self.assertIn(plan2.root, plan.nodes)
        self.assertEqual(plan, plan2.root.plan)
        self.assertIn((mount_node, plan2.root), plan.edges)
        self.assertEqual(len(plan.edges), 2)
        self.assertEqual(len(plan.nodes), 3)


class TestPlanNode(unittest.TestCase):
    def test_plan_node_creation(self):
        node = PlanNode()
        self.assertIsInstance(node, PlanNode)
        self.assertEqual(node.status, TaskStatus.CREATED)
        self.assertEqual(node.plan, None)
        self.assertLessEqual(node.start_time, datetime.now())

    def test_plan_node_parent(self):
        node = PlanNode()
        plan = Plan(node)
        node2 = PlanNode()
        plan.add_edge(node, node2)

        self.assertIsNone(node.parent)
        self.assertEqual(node, node2.parent)

    def test_plan_all_parents(self):
        node = PlanNode()
        plan = Plan(node)
        node2 = PlanNode()
        plan.add_edge(node, node2)
        node3 = PlanNode()
        plan.add_edge(node2, node3)

        self.assertEqual(node.all_parents, [])
        self.assertEqual(node2.all_parents, [node])
        self.assertEqual(node3.all_parents, [node2, node])

    def test_plan_node_children(self):
        node = PlanNode()
        plan = Plan(node)

        self.assertEqual([], node.children)

        node2 = PlanNode()
        plan.add_edge(node, node2)
        self.assertEqual([node2], node.children)

        node3 = PlanNode()
        plan.add_edge(node, node3)
        self.assertEqual([node2, node3], node.children)

    def test_plan_node_recursive_children(self):
        node = PlanNode()
        plan = Plan(node)

        self.assertEqual([], node.recursive_children)

        node2 = PlanNode()
        plan.add_edge(node, node2)
        self.assertEqual([node2], node.recursive_children)

        node3 = PlanNode()
        plan.add_edge(node2, node3)
        self.assertEqual([node2, node3], node.recursive_children)

    def test_plan_node_is_leaf(self):
        node = PlanNode()
        plan = Plan(node)
        node2 = PlanNode()
        plan.add_edge(node, node2)

        self.assertFalse(node.is_leaf)
        self.assertTrue(node2.is_leaf)

    def test_plan_node_subtree(self):
        node = PlanNode()
        node2 = PlanNode()
        node3 = PlanNode()
        plan = Plan(node)
        plan.add_edge(node, node2)
        plan.add_edge(node2, node3)

        sub_tree = node2.subtree
        self.assertEqual(node2, sub_tree.root)
        self.assertIn(node2, plan.nodes)
        self.assertIn(node3, plan.nodes)
        self.assertEqual(len(sub_tree.edges), 1)
        self.assertIn((node2, node3), sub_tree.edges)

class TestPlanInterrupt(BulletWorldTestCase):
        def test_interrupt_plan(self):

            def interrupt_plan():
                Plan.current_plan.root.interrupt()

            code_node = CodeNode(interrupt_plan)
            with simulated_robot:
                SequentialPlan(MoveTorsoActionDescription(TorsoState.HIGH), Plan(code_node), MoveTorsoActionDescription([TorsoState.LOW])).perform()

            self.assertEqual(0.3, self.robot.joints["torso_lift_joint"].position)

        @unittest.skip(
            "There is some weird error here that causes the interpreter to abort with exit code 134, something with thread handling. Needs more investigation")
        def test_pause_plan(self):
            def node_sleep():
                sleep(1)

            def pause_plan():
                Plan.current_plan.root.pause()
                self.assertEqual(0, self.robot.joints["torso_lift_joint"].position)
                Plan.current_plan.root.resume()
                sleep(3)
                self.assertEqual(0.3, self.robot.joints["torso_lift_joint"].position)

            code_node = CodeNode(pause_plan)
            sleep_node = CodeNode(node_sleep)
            robot_plan = SequentialPlan(Plan(sleep_node), MoveTorsoActionDescription([TorsoState.HIGH]))

            with simulated_robot:
                ParallelPlan(Plan(code_node), robot_plan).perform()

            self.assertEqual(0.3, self.robot.joints["torso_lift_joint"].position)


class AlgebraTest(BulletWorldTestCase):

    def test_algebra(self):
        sp = SequentialPlan(MoveTorsoActionDescription(None),
                            NavigateActionDescription(None),
                            MoveTorsoActionDescription(None))

        p = Parameterizer(sp)
        distribution = p.create_fully_factorized_distribution()

        conditions = []
        for state in TorsoState:
            v1 = p.get_variable("MoveTorsoAction_0.torso_state")
            v2 = p.get_variable("MoveTorsoAction_2.torso_state")
            se = SimpleEvent({v1: state, v2: state})
            conditions.append(se)

        condition = Event(*conditions)
        condition.fill_missing_variables(p.variables)

        navigate_condition = SimpleEvent({
            p.get_variable("NavigateAction_1.target_location.pose.position.z"): 0,
            p.get_variable("NavigateAction_1.target_location.pose.orientation.x"): 0,
            p.get_variable("NavigateAction_1.target_location.pose.orientation.y"): 0,
            p.get_variable("NavigateAction_1.target_location.pose.orientation.z"): 0,
            p.get_variable("NavigateAction_1.target_location.pose.orientation.w"): 1
        })
        navigate_condition.fill_missing_variables(p.variables)
        condition &= navigate_condition.as_composite_set()

        condition &= p.create_restrictions().as_composite_set()

        conditional, p_c = distribution.conditional(condition)

        for i in range(10):
            sample = conditional.sample(1)

            resolved = p.plan_from_sample(conditional, sample[0])
            with simulated_robot:
                resolved.perform()



from .default_process_modules import *
from ..worlds.bullet_world import World
from ..process_module import ProcessModuleManager
from ..robot_description import RobotDescription
from ..datastructures.enums import Arms, ExecutionType, Grasp, StaticJointState
from .default_process_modules import _move_arm_tcp


class DonbotMoveHead(DefaultMoveHead):
    """
    Moves the head of the iai_donbot robot to look at a specified point in the world coordinate frame.
    This point can be a position or an object, and the orientation is calculated based on the
    robot's base and camera alignment.
    """

    def _execute(self, desig):
        target = desig.target.copy()
        robot = World.robot
        local_transformer = LocalTransformer()

        # Rotate the arm to have a nice starting seed for the IK. If we dont do this, we regularly run into
        # self-collision in the simulation. Another solution would be to look into (possibly collision gradient
        # based) collision avoidance for pinocchio.
        perceive_state = RobotDescription.current_robot_description.get_static_joint_chain("left_arm",
                                                                                           "looking")
        robot.set_multiple_joint_positions(perceive_state)
        pose_in_pan = local_transformer.transform_pose(target,
                                                        robot.get_link_tf_frame("ur5_shoulder_link")).position_as_list()
        new_pan = np.arctan2(pose_in_pan[1], pose_in_pan[0])
        current_pan = robot.get_joint_position("ur5_shoulder_pan_joint")
        robot.set_joint_position("ur5_shoulder_pan_joint", new_pan + current_pan)

        # Compute the position the camera should take when looking. Due to the geometry of the arm and the placements of
        # the links, we translate the shoulder link pose by 20cm in -y direction, since this is roughly aligned with the
        # gripper when donbot is in the parking pose. The z-position is currently hardcoded to a moderate height, a bit
        # higher and lower should still work just as well. Could in theory be dynamically set depending on the required
        # camera height to see an object, similar to setting up a torso to be able to see a certain object.
        base_frame_pose: Pose = robot.get_link_pose("ur5_shoulder_link").copy()
        base_frame_pose = local_transformer.translate_pose_along_local_axis(base_frame_pose, [0, -1, 0], 0.2)
        base_frame_pose.position.z += 0.6
        base_position = np.array(base_frame_pose.position_as_list())

        # Define axis directly pointing toward the target
        target_position = np.array(target.position_as_list())
        direction_vector = target_position - base_position
        direction_vector /= np.linalg.norm(direction_vector)

        # Define that z-axis denotes the world's "up" direction
        world_up_z = np.array([0, 0, 1])

        # Define y-axis perpendicular to X and Z, according to right-hand rule
        right_hand_y = np.cross(world_up_z, direction_vector)
        right_hand_y /= np.linalg.norm(right_hand_y)

        # Recompute Z-axis to disallow roll (since we want the camera to be somewhat level if possible)
        corrected_up_z = np.cross(direction_vector, right_hand_y)

        # Construct rotation matrix
        rotation_matrix = np.column_stack((direction_vector, right_hand_y, corrected_up_z))

        # Convert to quaternion, and construct pose
        orientation_quat = R.from_matrix(rotation_matrix).as_quat()
        adjusted_pose = Pose(base_position.tolist(), orientation_quat.tolist())

        # Get front grasp of gripper, apply it to the pose, and then roll 180°. This works for donbot, because
        # the camera has the same frame orientation, albeit, rotated by 180° around z (which is the "forward" axis of the frame)
        side_grasp, top_grasp, horizontal = (Grasp.FRONT, None, False)
        grasp_orientation = RobotDescription.current_robot_description.get_arm_chain(Arms.LEFT).end_effector.get_grasp(
            side_grasp, top_grasp, horizontal)
        adjusted_pose.rotate_by_quaternion(grasp_orientation)
        adjusted_pose.rotate_by_quaternion([0, 0, 1, 0])

        # move the camera link to the computed pose
        camera_link = RobotDescription.current_robot_description.get_camera_link()
        _move_arm_tcp(adjusted_pose, robot, Arms.LEFT, camera_link)


class DonbotManager(DefaultManager):

    def __init__(self):
        super().__init__()
        self.robot_name = "iai_donbot"

    def looking(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DonbotMoveHead(self._looking_lock)

DonbotManager()



import datetime

from typing_extensions import Type

from .world_conf import WorldConfig
from pycram.description import ObjectDescription
from pycram.helper import find_multiverse_resources_path
from pycram.object_descriptors.mjcf import ObjectDescription as MJCF


class MultiverseConfig(WorldConfig):
    # Multiverse Configuration
    resources_path = find_multiverse_resources_path()
    """
    The path to the Multiverse resources directory.
    """

    # Multiverse Socket Configuration
    HOST: str = "tcp://127.0.0.1"
    SERVER_HOST: str = HOST
    SERVER_PORT: str = 7000
    BASE_CLIENT_PORT: int = 9000

    # Multiverse Client Configuration
    READER_MAX_WAIT_TIME_FOR_DATA: datetime.timedelta = datetime.timedelta(milliseconds=1000)
    """
    The maximum wait time for the data in seconds.
    """

    # Multiverse Simulation Configuration
    simulation_time_step: datetime.timedelta = datetime.timedelta(milliseconds=10)
    simulation_frequency: int = int(1 / simulation_time_step.total_seconds())
    """
    The time step of the simulation in seconds and the frequency of the simulation in Hz.
    """

    simulation_wait_time_factor: float = 1.0
    """
    The factor to multiply the simulation wait time with, this is used to adjust the simulation wait time to account for
    the time taken by the simulation to process the request, this depends on the computational power of the machine
    running the simulation.
    """

    use_static_mode: bool = True
    """
    If True, the simulation will always be in paused state unless the simulate() function is called, this behaves 
    similar to bullet_world which uses the bullet physics engine.
    """

    use_controller: bool = True
    """
    Only used when use_static_mode is False. This turns on the controller for the robot joints.
    """

    default_description_type: Type[ObjectDescription] = MJCF
    """
    The default description type for the objects.
    """

    use_physics_simulator_state: bool = False
    """
    Whether to use the physics simulator state when restoring or saving the world state.
    """

    validate_goals = True

    clear_cache_at_start = True

    let_pycram_move_attached_objects = False
    let_pycram_handle_spawning = False

    position_tolerance = 2e-2
    prismatic_joint_position_tolerance = 2e-2

    use_giskard_monitor = False
    allow_gripper_collision = True

    use_multiverse_process_modules = True

    depth_images_are_in_meter = True

    max_batch_size_for_rays = None
